{ File   : ISN2WAV.dpr
  Author : Deniz Oezmen
  Created: 2005-10-23
  Changed: 2007-08-05

  The Longest Journey ISN converter.
}
program ISN2WAV;

{$APPTYPE CONSOLE}

uses
  SysUtils, StrUtils;

type
  TBuffer  = array of Byte;

var
  ISNFile      : file;
  WAVHeader    : packed array[0..43] of Char =
    ( 'R',  'I',  'F',  'F', #$00, #$00, #$00, #$00,  'W',  'A',  'V',  'E',
      'f',  'm',  't',  ' ', #$10, #$00, #$00, #$00, #$01, #$00, #$01, #$00,
     #$22, #$56, #$00, #$00, #$22, #$56, #$00, #$00, #$02, #$00, #$10, #$00,
     'd',  'a',  't',  'a',  #$00, #$00, #$00, #$00);

function Prefix(FileName: string): string;
var
  i: Longint;
begin
  FileName := ExpandFileName(FileName);

  i := Length(FileName) + 1;
  repeat
    Dec(i)
  until (i = 0) or (PosEx('.', FileName, i) > 0);

  if (i = 0) or (PosEx('\', FileName, i + 1) > 0) then
    Result := FileName
  else
    Result := Copy(FileName, 1, i - 1)
end;

function OpenISNFile(var ISNFile: file; ISNName: string): Boolean;
begin
  FileMode := fmShareDenyWrite;
  AssignFile(ISNFile, ISNName);

  {$I-}
  Reset(ISNFile, 1);
  {$I+}
  OpenISNFile := (IOResult = 0)
end;

function ReadStrFromISN: string;
var
  c: Char;
begin
  c := #000;
  Result := '';

  while (not Eof(ISNFile)) and (Length(Result) < 255) do
  begin
    BlockRead(ISNFile, c, 1);
    if c = #032 then
      Break
    else if c in ['a'..'z', 'A'..'Z', '0'..'9', '_', '-', #000, '.'] then
      Result := Result + c
    else
    begin
      Result := '';
      Break
    end
  end;

  if Eof(ISNFile) or (Length(Result) = 255) then
    Result := ''
end;

procedure IMA_GetNewSample(var Sample: Smallint; var Idx: Byte;
  SampleCode: Byte);
const
  StepTab: array[0..88] of Smallint =
    (    7,     8,     9,    10,    11,    12,    13,    14,
        16,    17,    19,    21,    23,    25,    28,    31,
        34,    37,    41,    45,    50,    55,    60,    66,
        73,    80,    88,    97,   107,   118,   130,   143,
       157,   173,   190,   209,   230,   253,   279,   307,
       337,   371,   408,   449,   494,   544,   598,   658,
       724,   796,   876,   963,  1060,  1166,  1282,  1411,
      1552,  1707,  1878,  2066,  2272,  2499,  2749,  3024,
      3327,  3660,  4026,  4428,  4871,  5358,  5894,  6484,
      7132,  7845,  8630,  9493, 10442, 11487, 12635, 13899,
     15289, 16818, 18500, 20350, 22385, 24623, 27086, 29794,
     32767);
  IndexTab: array[0..15] of Smallint =
    ( -1, -1, -1, -1, 2, 4, 6, 8,-1, -1, -1, -1, 2, 4, 6, 8 );
var
  Samp0,
  Idx0,
  SampX,
  Diff : Longint;
begin
  // expand parameters' value range for computations
  Samp0 := Longint(Sample);
  Idx0  := Longint(Idx);
  SampX := Longint(SampleCode);

  // calculate difference
  Diff  := 0;
  if SampX and $04 <> 0 then
    Diff := Diff + StepTab[Idx0];
  if SampX and $02 <> 0 then
    Diff := Diff + (StepTab[Idx0] shr 1);
  if SampX and $01 <> 0 then
    Diff := Diff + (StepTab[Idx0] shr 2);
  Diff := Diff + (StepTab[Idx0] shr 3);
  if SampX and $08 <> 0 then
    Diff := -Diff;

  // calculate new sample
  Samp0 := Samp0 + Diff;
  if Samp0 > 32767 then
    Samp0 := 32767
  else if Samp0 < -32768 then
    Samp0 := -32768;

  // calculate new step table index
  Idx0 := Idx0 + IndexTab[SampX];
  if Idx0 > 88 then
    Idx0 := 88
  else if Idx0 < 0 then
    Idx0 := 0;

  // write values back to parameters
  Sample := Smallint(Samp0);
  Idx    := Byte(Idx0)
end;

procedure IMA_ConvertBlock(BlockSize, Channels: Longword;
  InBuffer: array of Byte; var OutBuffer: array of Smallint);
var
  i,
  InPos,
  OutPos,
  Channel: Longword;
  Samples: array of Smallint;
  Indexes: array of Byte;
begin
  SetLength(Samples, Channels);
  SetLength(Indexes, Channels);

  // initialize samples and indexes
  for i := 0 to Channels - 1 do
  begin
    Move(InBuffer[i * 4], Samples[i], 2);
    Indexes[i] := InBuffer[i * 4 + 2];
    OutBuffer[i] := Samples[i]
  end;

  // convert rest of block
  InPos   := Channels * 4;
  OutPos  := Channels;
  Channel := Channels - 1;

  // since we're decoding 4-bit ADPCM, look at each byte twice
  for i := 1 to (BlockSize - Channels * 4) * 2 do
  begin
    // lower nibble: high channel
    if i mod 2 = 1 then
      IMA_GetNewSample(Samples[Channel], Indexes[Channel],
        InBuffer[InPos] and $0f)
    else
      IMA_GetNewSample(Samples[Channel], Indexes[Channel],
        InBuffer[InPos] shr 4);

    OutBuffer[OutPos + Channel] := Samples[Channel];

    // proceed to next channel
    if Channel > 0 then
      Dec(Channel)
    else
    begin
      Channel := Channels - 1;
      Inc(OutPos, Channels)
    end;

    if i mod 2 = 0 then
      Inc(InPos)
  end;

  Finalize(Samples);
  Finalize(Indexes)
end;

procedure Convert_IMA_ADPCM(ISNName: string);
var
  Channels,
  BitsPerSample,
  BlockAlign,
  BlockSize,
  SampPerBlock  : Word;
  i,
  Frequency,
  BytesPerSecond,
  NumSamples,
  Size,
  UncompSize    : Longword;
  InBuffer      : TBuffer;
  OutBuffer     : array of Smallint;
  WAVFile       : file;
begin
  ISNName := Prefix(ISNName) + '.wav';
  WriteLn(' => ', ISNName);

  // read ISN header
  BlockSize := StrToInt(ReadStrFromISN);
  ReadStrFromISN;
  ReadStrFromISN;
  Channels := StrToInt(ReadStrFromISN) + 1;
  ReadStrFromISN;
  Frequency := 44100 div StrToInt(ReadStrFromISN);
  ReadStrFromISN;
  ReadStrFromISN;
  Size := StrToInt(ReadStrFromISN);

  // calculate remaining 4-bit IMA_ADPCM parameters
  SampPerBlock := (((BlockSize - (4 * Channels)) * 8) div (4 * Channels)) + 1;
  NumSamples := (Size div BlockSize) * SampPerBlock;
  if Size mod BlockSize > 0 then
    NumSamples := NumSamples + ((Size mod BlockSize) - 4 * Channels) * 2 +
      Channels;

  // set remaining WAVE parameters
  UncompSize := NumSamples * Channels * 2;
  BytesPerSecond := Frequency * Channels * 2;
  BlockAlign := Channels * 2;
  BitsPerSample := 16;

  Inc(UncompSize, 36);

  Move(UncompSize, WAVHeader[4], 4);
  Move(Channels, WAVHeader[22], 2);
  Move(Frequency, WAVHeader[24], 4);
  Move(BytesPerSecond, WAVHeader[28], 4);
  Move(BlockAlign, WAVHeader[32], 2);
  Move(BitsPerSample, WAVHeader[34], 2);

  Dec(UncompSize, 36);

  Move(UncompSize, WAVHeader[40], 4);

  FileMode := fmOpenReadWrite;
  Assign(WAVFile, ISNName);

  {$I-}
  Rewrite(WAVFile, 1);
  {$I+}
  if IOResult <> 0 then
  begin
    WriteLn('  Write error.');
    Exit
  end;

  BlockWrite(WAVFile, WAVHeader[0], SizeOf(WAVHeader));

  SetLength(InBuffer, BlockSize);
  SetLength(OutBuffer, SampPerBlock * Channels);

  for i := 1 to Size div BlockSize do
  begin
    BlockRead(ISNFile, InBuffer[0], BlockSize);
    IMA_ConvertBlock(BlockSize, Channels, InBuffer, OutBuffer);
    BlockWrite(WAVFile, OutBuffer[0], SampPerBlock * Channels * 2)
  end;
  if Size mod BlockSize <> 0 then
  begin
    // does this situation ever occur?
    NumSamples := ((Size mod BlockSize) - 4 * Channels) * 2 + Channels;
    BlockSize := Size mod BlockSize;
    BlockRead(ISNFile, InBuffer[0], BlockSize);
    IMA_ConvertBlock(BlockSize, 2, InBuffer, OutBuffer);
    BlockWrite(WAVFile, OutBuffer[0], NumSamples * Channels * 2)
  end;

  Finalize(InBuffer);
  Finalize(OutBuffer);

  Close(WAVFile)
end;

procedure Convert_Sound(ISNName: string);
var
  Channels,
  BitsPerSample,
  BlockAlign    : Word;
  Frequency,
  BytesPerSecond,
  Size          : Longword;
  Buffer        : TBuffer;
  WAVFile       : file;
begin
  ISNName := Prefix(ISNName) + '.wav';
  WriteLn(' => ', ISNName);

  // read SSN header
  ReadStrFromISN;
  Size := StrToInt(ReadStrFromISN);
  Channels := StrToInt(ReadStrFromISN) + 1;
  ReadStrFromISN;
  Frequency := 44100 div StrToInt(ReadStrFromISN);
  ReadStrFromISN;
  ReadStrFromISN;

  // set remaining WAVE parameters
  BytesPerSecond := Frequency * Channels * 2;
  BlockAlign := Channels * 2;
  BitsPerSample := 16;

  Inc(Size, 36);

  Move(Size, WAVHeader[4], 4);
  Move(Channels, WAVHeader[22], 2);
  Move(Frequency, WAVHeader[24], 4);
  Move(BytesPerSecond, WAVHeader[28], 4);
  Move(BlockAlign, WAVHeader[32], 2);
  Move(BitsPerSample, WAVHeader[34], 2);

  Dec(Size, 36);

  Move(Size, WAVHeader[40], 4);

  FileMode := fmOpenReadWrite;
  Assign(WAVFile, ISNName);

  {$I-}
  Rewrite(WAVFile, 1);
  {$I+}
  if IOResult <> 0 then
  begin
    WriteLn('  Write error.');
    Exit
  end;

  BlockWrite(WAVFile, WAVHeader[0], SizeOf(WAVHeader));

  SetLength(Buffer, Size);
  BlockRead(ISNFile, Buffer[0], Size);
  BlockWrite(WAVFile, Buffer[0], Size);

  Finalize(Buffer);

  Close(WAVFile)
end;

procedure Convert(ISNName: string);
var
  Codec: string;
begin
  Write(ISNName);

  Codec := ReadStrFromISN;

  if Codec = 'Sound' then
    Convert_Sound(ISNName)
  else if Codec = 'IMA_ADPCM_Sound' then
    Convert_IMA_ADPCM(ISNName)
  else if Codec = '' then
    WriteLn(' is not an ISN/ISS/SSN file.')
  else
    WriteLn(': unsupported codec "', Codec, '".')
end;

procedure ScanFiles(FileParam: string);
var
  FileName,
  Dir,
  ISNName  : string;
  SearchRec: TSearchRec;
begin
  // get the directory information from FileParam
  Dir := ExtractFilePath(FileParam);

  // search for all files matching FileParam
  if FindFirst(FileParam, faAnyFile - faDirectory, SearchRec) = 0 then
  begin
    repeat
      FileName := SearchRec.Name;
      ISNName := Dir + FileName;

      // if this file is accessible, analyze it
      if OpenISNFile(ISNFile, ISNName) then
      begin
        Convert(ISNName);
        Close(ISNFile)
      end
      else
        Writeln(ISNName, ' not accessible')
    until FindNext(SearchRec) <> 0
  end
  else
    WriteLn(Format('No matches for "%s"', [FileParam]));

  FindClose(SearchRec)
end;

begin
  if ParamCount >= 1 then
    ScanFiles(ParamStr(1))
  else
    Writeln('Incorrect number of parameters.')
end.
