{ File    : $URL: file:///D:/SVN/svnDelphi/trunk/TLJ/TLJView/src/backend/Animation.pas $
  Author  : $Author: Deniz Oezmen $
  Changed : $Date: 2011-02-06 01:15:38 +0100 (So, 06 Feb 2011) $
  Revision: $Rev: 4 $

  CIR/TM/ANI animation wrapper.
}
unit Animation;

interface

uses
  GLScene, Graphics, Classes;

type
  TModel = record
             Header   : record
                          ID       : Longint;
                          Version  : Longint;
                          Unknown1 : Longint;
                          Unknown2 : Longint;
                          Unknown3 : Single;
                        end;
             Materials: array of record
                          Name       : string;
                          Unknown1   : Longint;
                          TextureName: string;
                          ColourR    : Single;
                          ColourG    : Single;
                          ColourB    : Single;
                        end;
             Unknown4 : array of packed record
                          Unknown4_1: Single;
                          Unknown4_2: Single;
                          Unknown4_3: Single;
                          Unknown4_4: Single;
                        end;
             Skeleton : array of record
                          Name    : string;
                          Unknown1: Single;
                          Children: array of Longint;
                        end;
             Groups   : array of record
                          Name    : string;
                          Faces   : array of record
                                      MaterialIndex: Longint;
                                      Vertices     : array of packed record
                                                       PosX1     : Single;
                                                       PosY1     : Single;
                                                       PosZ1     : Single;
                                                       PosX2     : Single;
                                                       PosY2     : Single;
                                                       PosZ2     : Single;
                                                       NormalX   : Single;
                                                       NormalY   : Single;
                                                       NormalZ   : Single;
                                                       TextureS  : Single;
                                                       TextureT  : Single;
                                                       BoneIndex1: Longint;
                                                       BoneIndex2: Longint;
                                                       BoneWeight: Single;
                                                     end;
                                      Triangles    : array of packed record
                                                       VertexIndex1: Longint;
                                                       VertexIndex2: Longint;
                                                       VertexIndex3: Longint;
                                                     end;
                                    end;
                          Unknown1: array of packed record
                                      Unknown1_1: Single;
                                      Unknown1_2: Single;
                                      Unknown1_3: Single;
                                      Unknown1_4: Single;
                                      Unknown1_5: Longint;
                                    end;
                          Unknown2: array of packed record
                                      Unknown2_01: string;
                                      Unknown2_02: Single;
                                      Unknown2_03: Single;
                                      Unknown2_04: Single;
                                      Unknown2_05: Single;
                                      Unknown2_06: Single;
                                      Unknown2_07: Single;
                                      Unknown2_08: Single;
                                      Unknown2_09: Single;
                                      Unknown2_10: Longint;
                                    end;
                        end;

           end;

  TIntAnimation = record
                    Header   : record
                                 ID: Longint;
                                 Version: Longint;
                                 Unknown1: Longint;
                                 Unknown2: Longint;
                                 MaxTime : Longint;
                               end;
                    BoneAnims: array of record
                                 BoneIndex: Longint;
                                 KeyDatas : array of packed record
                                              AnimTime: Longint;
                                              QRotX   : Single;
                                              QRotY   : Single;
                                              QRotZ   : Single;
                                              QRotW   : Single;
                                              PosX    : Single;
                                              PosY    : Single;
                                              PosZ    : Single;
                                            end;
                               end;
                  end;

procedure InitScene(const Scene: TGLScene);
procedure DeinitScene;
procedure AssignAnimation(const IntAnimation: TIntAnimation);
procedure AssignModel(const Model: TModel);
procedure AssignTextureLib(const Images: array of TBitmap;
  const Names: TStringList);
procedure EraseAnimation(var IntAnimation: TIntAnimation); overload;
procedure EraseAnimation; overload;
procedure EraseModel(var Model: TModel); overload;
procedure EraseModel; overload;
procedure EraseTextureLib;
function UpdateAnimation(FrameTime: Double; SmoothWrap: Boolean): Double;
function IsAnimationAssigned: Boolean;

implementation

uses
  GLVectorFileObjects, GLTexture, VectorGeometry, GLObjects, SysUtils, Types,
  Math;

const
  TextureAmbient = $3b2321;
  TextureDiffuse = $92d8fe;
  LightAmbient   = $b67e78;
  LightDiffuse   = $fffcf0;

type
  TTexture = record
    Name  : string;
    Bitmap: TBitmap;
  end;

  TTextureLib = array of TTexture;

var
  CurScene : TGLScene = nil;
  CurActor : TGLActor = nil;
  CurAnim  : TIntAnimation;
  CurModel : TModel;
  CurTexLib: TTextureLib;

(* internal procedures ********************************************************)

procedure InitModel;
var
  i, j,
  VertCtr,
  BoneCtr  : Longint;
  Bone     : TSkeletonBone;
  Frame    : TSkeletonFrame;
  Bones    : TVertexBoneWeightDynArray;
  Mesh     : TSkeletonMeshObject;
  FaceGroup: TFGVertexNormalTexIndexList;
begin
  if CurActor = nil then
    Exit;

  // skeleton setup
  CurActor.Skeleton.Clear;
  Bone := TSkeletonBone.CreateOwned(CurActor.Skeleton.RootBones);
  Bone.BoneID := 0;

  for i := 0 to High(CurModel.Skeleton) do
    with CurModel.Skeleton[i] do
    begin
      Bone := CurActor.Skeleton.RootBones.BoneByID(i);
      Bone.Name := Name;

      for j := 0 to High(Children) do
        TSkeletonBone.CreateOwned(Bone).BoneID := Children[j]
    end;

  // set empty anim position while animation is not yet inited
  CurActor.Skeleton.Frames.Clear;
  Frame := TSkeletonFrame.CreateOwned(CurActor.Skeleton.Frames);
  Frame.TransformMode := sftQuaternion;

  for i := 0 to High(CurModel.Skeleton) do
  begin
    Frame.Position.Add(NullVector);
    Frame.Quaternion.Add(IdentityQuaternion)
  end;

  SetLength(Bones, 2);

  // mesh setup
  CurActor.MeshObjects.Clear;
  Mesh := TSkeletonMeshObject.CreateOwned(CurActor.MeshObjects);
  Mesh.BonesPerVertex := 2;
  Mesh.Mode := momFaceGroups;

  VertCtr := 0;
  BoneCtr := Length(CurModel.Skeleton);

  // load faces for first group
  for i := 0 to High(CurModel.Groups[0].Faces) do
    with CurModel.Groups[0].Faces[i] do
    begin
      // setup vertices
      for j := 0 to High(Vertices) do
        with Vertices[j] do
        begin
          // Ugly hack: GLScene only takes one position per vertex, but we have
          // each vertex expressed in two bone systems. "Solution": Add "pseudo
          // bones", which then act as a reference system for the vertex
          // positions.
          // This alters the skeleton and should be reworked ...
          Bone := CurActor.Skeleton.RootBones.BoneByID(BoneIndex1);
          TSkeletonBone.CreateOwned(Bone).BoneId := BoneCtr;
          Inc(BoneCtr);
          with CurActor.Skeleton.Frames[0] do
          begin
            Position.Add(AffineVectorMake(PosX1, PosY1, PosZ1));
            Quaternion.Add(IdentityQuaternion)
          end;
          Bone := CurActor.Skeleton.RootBones.BoneByID(BoneIndex2);
          TSkeletonBone.CreateOwned(Bone).BoneID := BoneCtr;
          Inc(BoneCtr);
          with CurActor.Skeleton.Frames[0] do
          begin
            Position.Add(AffineVectorMake(PosX2, PosY2, PosZ2));
            Quaternion.Add(IdentityQuaternion)
          end;
          Mesh.Vertices.Add(NullVector);
          Mesh.Normals.Add(NormalX, NormalY, NormalZ);
          Mesh.TexCoords.Add(TextureS, 1 - TextureT);
          Bones[0].BoneID := BoneCtr - 2;
          Bones[0].Weight := BoneWeight;
          Bones[1].BoneID := BoneCtr - 1;
          Bones[1].Weight := 1 - BoneWeight;
          Mesh.AddWeightedBones(Bones)
        end;

      FaceGroup := TFGVertexNormalTexIndexList.CreateOwned(Mesh.FaceGroups);
      FaceGroup.Mode := fgmmTriangles;
      FaceGroup.MaterialName := CurModel.Materials[MaterialIndex].Name;

      // setup triangles
      for j := 0 to High(Triangles) do
        with Triangles[j] do
        begin
          FaceGroup.VertexIndices.Add(VertCtr + VertexIndex1);
          FaceGroup.VertexIndices.Add(VertCtr + VertexIndex2);
          FaceGroup.VertexIndices.Add(VertCtr + VertexIndex3);
          FaceGroup.NormalIndices.Add(VertCtr + VertexIndex1);
          FaceGroup.NormalIndices.Add(VertCtr + VertexIndex2);
          FaceGroup.NormalIndices.Add(VertCtr + VertexIndex3);
          FaceGroup.TexCoordIndices.Add(VertCtr + VertexIndex1);
          FaceGroup.TexCoordIndices.Add(VertCtr + VertexIndex2);
          FaceGroup.TexCoordIndices.Add(VertCtr + VertexIndex3)
        end;

      Inc(VertCtr, Length(Vertices))
    end;

  // prepare preliminary mesh
  Mesh.PrepareBoneMatrixInvertedMeshes
end;

procedure InitTextureLib;
var
  i, j,
  ImgIdx  : Longint;
  Material: TGLLibMaterial;
  Bitmap  : TBitmap;
begin
  if CurActor = nil then
    Exit;

  if Assigned(CurActor.MaterialLibrary) then
    CurActor.MaterialLibrary.Free;

  CurActor.MaterialLibrary := TGLMaterialLibrary.Create(CurActor.Owner);
  CurActor.UseMeshMaterials := True;

  // fill material library
  for i := 0 to High(CurModel.Materials) do
    with CurModel.Materials[i] do
    begin
      // find texture by name
      ImgIdx := -1;
      for j := 0 to High(CurTexLib) do
        if AnsiLowerCase(CurTexLib[j].Name) = AnsiLowerCase(TextureName) then
          ImgIdx := j;

      // either load texture image ...
      if ImgIdx >= 0 then
        Bitmap := CurTexLib[ImgIdx].Bitmap
      else
      // ... or set solid colour
      begin
        Bitmap := TBitmap.Create;
        Bitmap.Height := 2;
        Bitmap.Width := 1;

        Bitmap.Canvas.Pixels[0, 0] := Ceil(ColourR * 255) or
          (Ceil(ColourG * 255) shl 8) or (Ceil(ColourB * 255) shl 16);
        Bitmap.Canvas.Pixels[0, 1] := Bitmap.Canvas.Pixels[0, 0]
      end;

      Material := CurActor.MaterialLibrary.AddTextureMaterial(Name, Bitmap);
      Material.Material.Texture.Enabled := True;
      Material.Material.Texture.TextureWrap := twBoth;
      Material.Material.Texture.TextureMode := tmModulate;
      Material.Material.FrontProperties.Ambient.AsWinColor := TextureAmbient;
      Material.Material.FrontProperties.Diffuse.AsWinColor := TextureDiffuse;
      Material.Material.BackProperties.Ambient.AsWinColor := TextureAmbient;
      Material.Material.BackProperties.Diffuse.AsWinColor := TextureDiffuse;

      if ImgIdx = -1 then
        Bitmap.Free
    end
end;

(* interface procedures *******************************************************)

procedure InitScene(const Scene: TGLScene);
var
  Light: TGLLightSource;
  DCube: TGLDummyCube;
begin
  CurScene := Scene;

  // general actor setup
  if Assigned(CurActor) then
  begin
    if Assigned(CurActor.MaterialLibrary) then
    begin
      CurActor.MaterialLibrary.Free;
      CurActor.MaterialLibrary := nil
    end;

    CurActor.Free;
    CurActor := nil
  end;

  CurScene.Objects.DeleteChildren;

  CurActor := TGLActor.CreateAsChild(CurScene.Objects);
  CurActor.Reference := aarSkeleton;
  CurActor.AnimationMode := aamLoop;

  // compensate for coordinate system handedness issues
  CurActor.ForceLocalMatrix(MakeReflectionMatrix(NullVector, ZVector));

  // material library setup
  CurActor.MaterialLibrary := TGLMaterialLibrary.Create(CurActor);
  CurActor.UseMeshMaterials := True;

  // rendering setup
  Light := TGLLightSource.CreateAsChild(CurScene.Objects);
  Light.Position.X := 100;
  Light.Position.Y := 100;
  Light.Position.Z := 100;
  Light.Ambient.AsWinColor := LightAmbient;
  Light.Diffuse.AsWinColor := LightDiffuse;
  Light.LightStyle := lsOmni;

  DCube := TGLDummyCube.CreateAsChild(CurScene.Objects);
  DCube.Position.X := 0;
  DCube.Position.Y := 90;
  DCube.Position.Z := 0
end;

procedure DeinitScene;
begin
  // destroy actor
  if Assigned(CurActor) then
  begin
    if Assigned(CurActor.MaterialLibrary) then
    begin
      CurActor.MaterialLibrary.Free;
      CurActor.MaterialLibrary := nil
    end;

    CurActor.Free;
    CurActor := nil
  end;

  // destroy scene
  if Assigned(CurScene) then
  begin
    CurScene.Objects.DeleteChildren;
    CurScene.Free
  end
end;

procedure AssignAnimation(const IntAnimation: TIntAnimation);
begin
  // clear current animation
  EraseAnimation;

  // assign new animation
  CurAnim := IntAnimation
end;

procedure AssignModel(const Model: TModel);
begin
  // clear current model
  EraseModel;

  // assign new model
  CurModel := Model;

  // initialize model
  InitModel;

  // init dependencies
  InitTextureLib
end;

procedure AssignTextureLib(const Images: array of TBitmap;
  const Names: TStringList);
var
  i,
  Len: Longint;
begin
  // clear current texture library
  EraseTextureLib;

  // load textures locally
  Len := Length(Images);
  if Names.Count < Len then
    Len := Names.Count;

  SetLength(CurTexLib, Len);

  for i := 0 to High(CurTexLib) do
    with CurTexLib[i] do
    begin
      Name := Names[i];
      Bitmap := TBitmap.Create;
      Bitmap.Assign(Images[i])
    end;

  // initialize texture use
  InitTextureLib

  // no dependent inits should be necessary ...
end;

procedure EraseAnimation(var IntAnimation: TIntAnimation); overload;
var
  i: Longint;
begin
  for i := 0 to High(IntAnimation.BoneAnims) do
    with IntAnimation.BoneAnims[i] do
    begin
      SetLength(KeyDatas, 0);
      KeyDatas := nil
    end;

  Finalize(IntAnimation.BoneAnims);
  SetLength(IntAnimation.BoneAnims, 0);
  IntAnimation.BoneAnims := nil
end;

procedure EraseAnimation; overload;
begin
  EraseAnimation(CurAnim)
end;

procedure EraseModel(var Model: TModel); overload;
var
  i, j: Longint;
begin
  Finalize(Model.Materials);
  SetLength(Model.Materials, 0);
  Model.Materials := nil;

  SetLength(Model.Unknown4, 0);
  Model.Unknown4 := nil;

  for i := 0 to High(Model.Skeleton) do
    with Model.Skeleton[i] do
    begin
      SetLength(Children, 0);
      Children := nil
    end;

  Finalize(Model.Skeleton);
  SetLength(Model.Skeleton, 0);
  Model.Skeleton := nil;

  for i := 0 to High(Model.Groups) do
    with Model.Groups[i] do
    begin
      for j := 0 to High(Faces) do
        with Faces[j] do
        begin
          SetLength(Vertices, 0);
          Vertices := nil;

          SetLength(Triangles, 0);
          Triangles := nil
        end;

      Finalize(Faces);
      SetLength(Faces, 0);
      Faces := nil;

      SetLength(Unknown1, 0);
      Unknown1 := nil;

      Finalize(Unknown2);
      SetLength(Unknown2, 0);
      Unknown2 := nil
    end;

  Finalize(Model.Groups);
  SetLength(Model.Groups, 0);
  Model.Groups := nil
end;

procedure EraseModel; overload;
begin
  EraseModel(CurModel)
end;

procedure EraseTextureLib;
var
  i: Longint;
begin
  for i := 0 to High(CurTexLib) do
    with CurTexLib[i] do
    begin
      SetLength(Name, 0);
      Bitmap.FreeImage;
      Bitmap.Free;
      Bitmap := nil
    end;

  Finalize(CurTexLib)
end;

function UpdateAnimation(FrameTime: Double; SmoothWrap: Boolean): Double;
const
  WrapTime = 1/30 * 1000;
var
  i, j,
  StartFrm,
  EndFrm  : Longint;
  AddTime,
  Delta   : Double;
  Quat1,
  Quat2   : TQuaternion;
begin
  Result := 0;

  if CurAnim.Header.MaxTime < 0 then
    Exit;
  if CurActor.Skeleton.Frames.Count <= 0 then
    Exit;

  if SmoothWrap then
    AddTime := WrapTime
  else
    AddTime := 0;

  FrameTime := FrameTime * 1000;

  if CurAnim.Header.MaxTime > 0 then
    while FrameTime > CurAnim.Header.MaxTime + AddTime do
      FrameTime := FrameTime - (CurAnim.Header.MaxTime + AddTime)
  else
    FrameTime := 0;

  Result := FrameTime / (CurAnim.Header.MaxTime + AddTime);

  for i := 0 to High(CurAnim.BoneAnims) do
    with CurAnim.BoneAnims[i] do
    begin
      if BoneIndex > CurActor.Skeleton.Frames[0].Quaternion.Count - 1 then
        Continue;
      if BoneIndex > CurActor.Skeleton.Frames[0].Position.Count - 1 then
        Continue;

      if High(KeyDatas) = 0 then
      begin
        StartFrm := 0;
        EndFrm := 0;
        Delta := 0
      end
      else if FrameTime <= CurAnim.Header.MaxTime then
      begin
        StartFrm := 0;
        EndFrm := High(KeyDatas);

        for j := 0 to High(KeyDatas) do
          if KeyDatas[j].AnimTime <= FrameTime then
            StartFrm := j
          else
            Break;
        for j := High(KeyDatas) downto 0 do
          if KeyDatas[j].AnimTime >= FrameTime then
            EndFrm := j
          else
            Break;

        if StartFrm <> EndFrm then
          Delta := (FrameTime - KeyDatas[StartFrm].AnimTime) /
            (KeyDatas[EndFrm].AnimTime - KeyDatas[StartFrm].AnimTime)
        else
          Delta := 0
      end
      else
      // never used if SmoothWrap = False
      begin
        StartFrm := High(KeyDatas);
        EndFrm := 0;

        Delta := (FrameTime - CurAnim.Header.MaxTime) / WrapTime
      end;

      Quat1 := QuaternionMake([KeyDatas[StartFrm].QRotX,
        KeyDatas[StartFrm].QRotY, KeyDatas[StartFrm].QRotZ],
        KeyDatas[StartFrm].QRotW);
      Quat2 := QuaternionMake([KeyDatas[EndFrm].QRotX,
        KeyDatas[EndFrm].QRotY, KeyDatas[EndFrm].QRotZ],
        KeyDatas[EndFrm].QRotW);
      NormalizeQuaternion(Quat1);
      NormalizeQuaternion(Quat2);

      CurActor.Skeleton.Frames[0].Quaternion[BoneIndex] := QuaternionSlerp(
        Quat1, Quat2, Delta);

      CurActor.Skeleton.Frames[0].Position[BoneIndex] := VectorLerp(
        AffineVectorMake(KeyDatas[StartFrm].PosX, KeyDatas[StartFrm].PosY,
        KeyDatas[StartFrm].PosZ), AffineVectorMake(KeyDatas[EndFrm].PosX,
        KeyDatas[EndFrm].PosY, KeyDatas[EndFrm].PosZ), Delta);
    end
end;

function IsAnimationAssigned: Boolean;
begin
  Result := Assigned(CurAnim.BoneAnims)
end;

initialization

finalization
  EraseAnimation;
  EraseModel;
  EraseTextureLib;

end.
