{ File   : Audiere.pas
  Author : Chad Austin, Jarrod Davis, Deniz Oezmen (see below for details)
  Created: 2007-02-22
  Changed: 2010-08-29

  Audiere Sound System 1.9.4 Delphi headers.

  Audiere Sound System is (c) Chad Austin.
  Original 1.9.2 Delphi conversion by Jarrod Davis.
  Interface changes and 1.9.4 adaptations by Deniz Oezmen.

  Beware: Adaptations have not been fully tested and might be wrong!

  Original copyright header follows.
}

//=========================================================================
// Audiere Sound System
// Version 1.9.2
// (c) 2002 Chad Austin
//
// This API uses principles explained at
// http://aegisknight.org/cppinterface.html
//
// This code licensed under the terms of the LGPL.  See the Audiere
// license.txt.
//-------------------------------------------------------------------------
// Delphi Conversion By:
// Jarrod Davis
// Jarrod Davis Software
// http://www.jarroddavis.com   - Jarrod Davis Software
// http://www.gamevisionsdk.com - Game Application Framework for Delphi
// support@jarroddavis.com      - Support Email
//-------------------------------------------------------------------------
// How to use:
//   * Include Audiere in your Uses statement
//   * Enable or Disable the DYNAMICS compiler define
//   * If Dynamic, be sure to call AdrLoadDLL before using any commands.
//     the DLL will be automaticlly unloaded at termination.
//   * If Static, then use as normal.
// History:
//   * Initial 1.9.2 release.
//   + Added dynamic loading. Use the DYNAMIC compiler define to control
//     this. When enabled you can then use ArdLoadLL/AdrUnloadDLL to
//     dynamiclly load/unload dll at runtime.
//=========================================================================

unit Audiere;

{$A+}
{$Z4}

{$DEFINE DYNAMIC}

interface

const
  DLL_NAME = 'audiere.dll';

type
  { TAdrSeekMode }
  TAdrSeekMode = (
    Adr_Seek_Begin,
    Adr_Seek_Current,
    Adr_Seek_End
  );

  { TAdrSoundEffectType }
  TAdrSoundEffectType = (
    Adr_SoundEffectType_Single,
    Adr_SoundEffectType_Multiple
  );

  { TAdrSampleFormat }
  TAdrSampleFormat = (
    Adr_SampleFormat_U8,
    Adr_SampleFormat_S16
  );

  { TAdrFileFormat }
  TAdrFileFormat = (
    FF_AUTODETECT,
    FF_WAV,
    FF_OGG,
    FF_FLAC,
    FF_MP3,
    FF_MOD,
    FF_AIFF,
    FF_SPEEX
  );

  { TAdrEventType }
  TAdrEventType = (
    ET_STOP
  );

  { TAdrReason }
  TAdrReason = (
    STOP_CALLED,
    STREAM_ENDED
  );

  { TAdrRefCounted  }
  TAdrRefCounted = class
  public
    procedure Ref;   virtual; stdcall; abstract;
    procedure UnRef; virtual; stdcall; abstract;
  end;

  { TAdrFile }
  TAdrFile = class(TAdrRefCounted)
  public
    function Read(aBuffer: Pointer; aSize: Integer): Integer; virtual; stdcall; abstract;
    function Seek(aPosition: Integer; aSeekMode: TAdrSeekMode): Boolean; virtual; stdcall; abstract;
    function Tell: Integer; virtual; stdcall; abstract;
  end;

  { TAdrSampleSource }
  TAdrSampleSource = class(TAdrRefCounted)
  public
    procedure GetFormat(var aChannelCount: Integer; var aSampleRate: Integer; var aSampleFormat: TAdrSampleFormat); virtual; stdcall; abstract;
    function  Read(aFrameCount: Integer; aBuffer: Pointer): Integer; virtual; stdcall; abstract;
    procedure Reset; virtual; stdcall; abstract;
    function  IsSeekable: Boolean; virtual; stdcall; abstract;
    function  GetLength: Integer; virtual; stdcall; abstract;
    procedure SetPosition(Position: Integer); virtual; stdcall; abstract;
    function  GetPosition: Integer; virtual; stdcall; abstract;
    function  GetRepeat: Boolean; virtual; stdcall; abstract;
    procedure SetRepeat(aRepeat: Boolean); virtual; stdcall; abstract;
    function  GetTagCount: Integer; virtual; stdcall; abstract;
    function  GetTagKey(i: Integer): PChar; virtual; stdcall; abstract;
    function  GetTagValue(i: Integer): PChar; virtual; stdcall; abstract;
    function  GetTagType(i: Integer): PChar; virtual; stdcall; abstract;
  end;

  { TAdrLoopPointSource }
  TAdrLoopPointSource = class(TAdrSampleSource)
  public
    procedure AddLoopPoint(aLocation, aTarget, aLoopCount: Integer); virtual; stdcall; abstract;
    procedure RemoveLoopPoint(aIndex: Integer); virtual; stdcall; abstract;
    function  GetLoopPointCount: Integer; virtual; stdcall; abstract;
    function  GetLoopPoint(aIndex: Integer; var aLocation, aTarget, aLoopCount: Integer): Boolean; virtual; stdcall; abstract;
  end;

  { TAdrOutputStream }
  TAdrOutputStream = class(TAdrRefCounted)
  public
    procedure Play; virtual; stdcall; abstract;
    procedure Stop; virtual; stdcall; abstract;
    function  IsPlaying: Boolean; virtual; stdcall; abstract;
    procedure Reset; virtual; stdcall; abstract;
    procedure SetRepeat(aRepeat: Boolean); virtual; stdcall; abstract;
    function  GetRepeat: Boolean; virtual; stdcall; abstract;
    procedure SetVolume(aVolume: Single); virtual; stdcall; abstract;
    function  GetVolume: Single; virtual; stdcall; abstract;
    procedure SetPan(aPan: Single); virtual; stdcall; abstract;
    function  GetPan: Single; virtual; stdcall; abstract;
    procedure SetPitchShift(aShift: Single); virtual; stdcall; abstract;
    function  GetPitchShift: Single; virtual; stdcall; abstract;
    function  IsSeekable: Boolean; virtual; stdcall; abstract;
    function  GetLength: Integer; virtual; stdcall; abstract;
    procedure SetPosition(aPosition: Integer); virtual; stdcall; abstract;
    function  GetPosition: Integer; virtual; stdcall; abstract;
  end;

  { TAdrEvent }
  TAdrEvent = class(TAdrRefCounted)
  public
    function  GetType: TAdrEventType; virtual; stdcall; abstract;
  end;

  { TAdrStopEvent }
  TAdrStopEvent = class(TAdrEvent)
  public
    function  GetType: TAdrEventType; override; stdcall;
    function  GetOutputStream: TAdrOutputStream; virtual; stdcall; abstract;
    function  GetReason: TAdrReason; virtual; stdcall; abstract;
  end;

  { TAdrCallback }
  TAdrCallback = class(TAdrRefCounted)
  public
    function  GetType: TAdrEventType; virtual; stdcall; abstract;
    procedure Call(aEvent: TAdrEvent); virtual; stdcall; abstract;
  end;

  { TAdrStopCallback }
  TAdrStopCallback = class(TAdrCallback)
  public
    function  GetType: TAdrEventType; override; stdcall;
    procedure Call(aEvent: TAdrEvent); override; stdcall;
    procedure StreamStopped(aStopEvent: TAdrEvent); virtual; stdcall; abstract;
  end;

  { TAdrAudioDevice }
  TAdrAudioDevice = class(TAdrRefCounted)
  public
    procedure Update; virtual; stdcall; abstract;
    function  OpenStream(aSource: TAdrSampleSource): TAdrOutputStream; virtual; stdcall; abstract;
    function  OpenBuffer(aSamples: Pointer; aFrameCount, aChannelCount, aSampleRate: Integer; aSampleFormat: TAdrSampleFormat): TAdrOutputStream; virtual; stdcall; abstract;
    function  GetName: PChar; virtual; stdcall; abstract;
    procedure RegisterCallback(Callback: TAdrCallback); virtual; stdcall; abstract;
    procedure UnregisterCallback(Callback: TAdrCallback); virtual; stdcall; abstract;
    procedure ClearCallbacks; virtual; stdcall; abstract;
  end;

  { TAdrSampleBuffer }
  TAdrSampleBuffer = class(TAdrRefCounted)
  public
    procedure GetFormat(var ChannelCount: Integer; var aSampleRate: Integer; var aSampleFormat: TAdrSampleFormat); virtual; stdcall; abstract;
    function  GetLength: Integer; virtual; stdcall; abstract;
    function  GetSamples: Pointer; virtual; stdcall; abstract;
    function  OpenStream: TAdrSampleSource; virtual; stdcall; abstract;
  end;

  { TAdrSoundEffect }
  TAdrSoundEffect = class(TAdrRefCounted)
  public
    procedure Play; virtual; stdcall; abstract;
    procedure Stop; virtual; stdcall; abstract;
    procedure SetVolume(aVolume: Single); virtual; stdcall; abstract;
    function  GetVolume: Single; virtual; stdcall; abstract;
    procedure SetPan(aPan: Single); virtual; stdcall; abstract;
    function  GetPan: Single; virtual; stdcall; abstract;
    procedure SetPitchShift(aShift: Single); virtual; stdcall; abstract;
    function  GetPitchShift: Single; virtual; stdcall; abstract;
  end;

  { TAdrCDDevice }
  TAdrCDDevice = class(TAdrRefCounted)
  protected
    constructor Create; virtual; abstract;
  public
    function  GetName: PChar; virtual; stdcall; abstract;
    function  GetTrackCount: Integer; virtual; stdcall; abstract;
    procedure Play(aTrack: Integer); virtual; stdcall; abstract;
    procedure Stop; virtual; stdcall; abstract;
    procedure Pause; virtual; stdcall; abstract;
    procedure Resume; virtual; stdcall; abstract;
    function  IsPlaying: Boolean; virtual; stdcall; abstract;
    function  ContainsCD: Boolean; virtual; stdcall; abstract;
    function  IsDoorOpen: Boolean; virtual; stdcall; abstract;
    procedure OpenDoor; virtual; stdcall; abstract;
    procedure CloseDoor; virtual; stdcall; abstract;
  end;

  { TAdrMIDIStream }
  TAdrMIDIStream = class(TAdrRefCounted)
  protected
    constructor Create; virtual; abstract;
  public
    procedure Play; virtual; stdcall; abstract;
    procedure Stop; virtual; stdcall; abstract;
    procedure Pause; virtual; stdcall; abstract;
    function  IsPlaying: Boolean; virtual; stdcall; abstract;
    function  GetLength: Integer; virtual; stdcall; abstract;
    function  GetPosition: Integer; virtual; stdcall; abstract;
    procedure SetPosition(aPosition: Integer); virtual; stdcall; abstract;
    function  GetRepeat: Boolean; virtual; stdcall; abstract;
    procedure SetRepeat(aRepeat: Boolean); virtual; stdcall; abstract;
  end;

  { TAdrMIDIDevice }
  TAdrMIDIDevice = class(TAdrRefCounted)
  protected
    constructor Create; virtual; abstract;
  public
    function  GetName: PChar; virtual; stdcall; abstract;
    function  OpenStream(aFilename: PChar): TAdrMIDIStream; virtual; stdcall; abstract;
  end;

  { TAdrStringVect }
  TAdrStringVect = array of string;

  { TAdrFileFormatDesc }
  TAdrFileFormatDesc = array of
    record
      Description: string;
      Extensions : TAdrStringVect;
    end;

  { TAdrAudioDeviceDesc }
  TAdrAudioDeviceDesc = array of
    record
      Name,
      Description: string;
    end;

{$IFDEF DYNAMIC}
function AdrLoadDLL(AdrLoadPath: AnsiString = ''): Boolean;
procedure AdrUnloadDLL; stdcall;
function AdrDLLLoaded: Boolean;
{$ENDIF}

function GetVersion: PChar;
procedure SplitString(var OutList: TAdrStringVect; InList: PChar; Delim: Char);
procedure GetSupportedFileFormats(var aFormats: TAdrFileFormatDesc);
procedure GetSupportedAudioDevices(var aDevices: TAdrAudioDeviceDesc);
function GetSampleSize(aFormat: TAdrSampleFormat): Integer;
function OpenDevice(aName, aParameters: PChar): TAdrAudioDevice;
function OpenSampleSource(aFilename: PChar; aFileFormat: TAdrFileFormat = FF_AUTODETECT): TAdrSampleSource; overload;
function OpenSampleSource(aFile: TAdrFile; aFileFormat: TAdrFileFormat = FF_AUTODETECT): TAdrSampleSource; overload;
function CreateTone(aFrequency: Double): TAdrSampleSource;
function CreateSquareWave(aFrequency: Double): TAdrSampleSource;
function CreateWhiteNoise: TAdrSampleSource;
function CreatePinkNoise: TAdrSampleSource;
function CreateLoopPointSource(aSource: TAdrSampleSource): TAdrLoopPointSource; overload;
function CreateLoopPointSource(aFilename: PChar; aFileFormat: TAdrFileFormat = FF_AUTODETECT): TAdrLoopPointSource; overload;
function CreateLoopPointSource(aFile: TAdrFile; aFileFormat: TAdrFileFormat = FF_AUTODETECT): TAdrLoopPointSource; overload;
function OpenSound(aDevice: TAdrAudioDevice; aSource: TAdrSampleSource; aStreaming: Boolean = False): TAdrOutputStream; overload;
function OpenSound(aDevice: TAdrAudioDevice; aFilename: PChar; aStreaming: Boolean = False; aFileFormat: TAdrFileFormat = FF_AUTODETECT): TAdrOutputStream; overload;
function OpenSound(aDevice: TAdrAudioDevice; aFile: TAdrFile; aStreaming: Boolean = False; aFileFormat: TAdrFileFormat = FF_AUTODETECT): TadrOutputStream; overload;
function CreateSampleBuffer(aSamples: Pointer; aFrameCount, aChannelCount, aSampleRate: Integer; aSampleFormat: TAdrSampleFormat): TAdrSampleBuffer; overload;
function CreateSampleBuffer(aSource: TAdrSampleSource): TAdrSampleBuffer; overload;
function OpenSoundEffect(aDevice: TAdrAudioDevice; aSource: TAdrSampleSource; aType: TAdrSoundEffectType): TAdrSoundEffect; overload;
function OpenSoundEffect(aDevice: TAdrAudioDevice; aFilename: PChar; aType: TAdrSoundEffectType; aFileFormat: TAdrFileFormat = FF_AUTODETECT): TAdrSoundEffect; overload;
function OpenSoundEffect(aDevice: TAdrAudioDevice; aFile: TAdrFile; aType: TAdrSoundEffectType; aFileFormat: TAdrFileFormat = FF_AUTODETECT): TAdrSoundEffect; overload;
function OpenFile(aFilename: PChar; aWriteable: Boolean): TAdrFile;
function CreateMemoryFile(aBuffer: Pointer; aSize: Integer): TAdrFile;
procedure EnumerateCDDevices(var aDevices: TAdrStringVect);
function OpenCDDevice(aDevice: PChar): TAdrCDDevice;
function OpenMIDIDevice(aDevice: PChar): TAdrMIDIDevice;

implementation

{$IFDEF DYNAMIC}
uses
  SysUtils, Windows;

var
  AdrDLL                         : HMODULE = 0;
  AdrGetVersion                  : function: PChar; stdcall = nil;
  AdrGetSupportedFileFormats     : function: PChar; stdcall = nil;
  AdrGetSupportedAudioDevices    : function: PChar; stdcall = nil;
  AdrGetSampleSize               : function(aFormat: TAdrSampleFormat): Integer; stdcall = nil;
  AdrOpenDevice                  : function(const aName: PChar; const aParams: PChar): TAdrAudioDevice; stdcall = nil;
  AdrOpenSampleSource            : function(const aFilename: PChar; aFileFormat: TAdrFileFormat): TAdrSampleSource; stdcall = nil;
  AdrOpenSampleSourceFromFile    : function(aFile: TAdrFile; aFileFormat: TAdrFileFormat): TAdrSampleSource; stdcall = nil;
  AdrCreateTone                  : function(aFrequency: Double): TAdrSampleSource; stdcall = nil;
  AdrCreateSquareWave            : function(aFrequency: Double): TAdrSampleSource; stdcall = nil;
  AdrCreateWhiteNoise            : function: TAdrSampleSource; stdcall = nil;
  AdrCreatePinkNoise             : function: TAdrSampleSource; stdcall = nil;
  AdrCreateLoopPointSource       : function(aSource: TAdrSampleSource): TAdrLoopPointSource; stdcall = nil;
  AdrOpenSound                   : function(aDevice: TAdrAudioDevice; aSource: TAdrSampleSource; aStreaming: LongBool): TAdrOutputStream; stdcall = nil;
  AdrCreateSampleBuffer          : function(aSamples: Pointer; aFrameCount, aChannelCount, aSampleRate: Integer; aSampleFormat: TAdrSampleFormat): TAdrSampleBuffer; stdcall = nil;
  AdrCreateSampleBufferFromSource: function(aSource: TAdrSampleSource): TAdrSampleBuffer; stdcall = nil;
  AdrOpenSoundEffect             : function(aDevice: TAdrAudioDevice; aSource: TAdrSampleSource; aType: TAdrSoundEffectType): TAdrSoundEffect; stdcall = nil;
  AdrOpenFile                    : function(aName: PChar; aWriteable: Boolean): TAdrFile; stdcall = nil;
  AdrCreateMemoryFile            : function(aBuffer: Pointer; aSize: Integer): TAdrFile; stdcall = nil;
  AdrEnumerateCDDevices          : function: PChar; stdcall = nil;
  AdrOpenCDDevice                : function(aName: PChar): TAdrCDDevice; stdcall = nil;
  AdrOpenMIDIDevice              : function(aname: PChar): TAdrMIDIDevice; stdcall = nil;

function AdrLoadDLL(AdrLoadPath: AnsiString = ''): Boolean;
begin
  Result := False;

  if AdrLoadPath = '' then
    AdrDLL := LoadLibrary(DLL_NAME)
  else
    AdrDLL := LoadLibrary(PAnsiChar(ExpandUNCFileName(
      IncludeTrailingPathDelimiter(AdrLoadPath) + DLL_NAME)));
  if AdrDLL = 0 then
    Exit;

  @AdrGetVersion                   := GetProcAddress(AdrDLL, '_AdrGetVersion@0');
  @AdrGetSupportedFileFormats      := GetProcAddress(AdrDLL, '_AdrGetSupportedFileFormats@0');
  @AdrGetSupportedAudioDevices     := GetProcAddress(AdrDLL, '_AdrGetSupportedAudioDevices@0');
  @AdrGetSampleSize                := GetProcAddress(AdrDLL, '_AdrGetSampleSize@4');
  @AdrOpenDevice                   := GetProcAddress(AdrDLL, '_AdrOpenDevice@8');
  @AdrOpenSampleSource             := GetProcAddress(AdrDLL, '_AdrOpenSampleSource@8');
  @AdrOpenSampleSourceFromFile     := GetProcAddress(AdrDLL, '_AdrOpenSampleSourceFromFile@8');
  @AdrCreateTone                   := GetProcAddress(AdrDLL, '_AdrCreateTone@8');
  @AdrCreateSquareWave             := GetProcAddress(AdrDLL, '_AdrCreateSquareWave@8');
  @AdrCreateWhiteNoise             := GetProcAddress(AdrDLL, '_AdrCreateWhiteNoise@0');
  @AdrCreatePinkNoise              := GetProcAddress(AdrDLL, '_AdrCreatePinkNoise@0');
  @AdrCreateLoopPointSource        := GetProcAddress(AdrDLL, '_AdrCreateLoopPointSource@4');
  @AdrOpenSound                    := GetProcAddress(AdrDLL, '_AdrOpenSound@12');
  @AdrCreateSampleBuffer           := GetProcAddress(AdrDLL, '_AdrCreateSampleBuffer@20');
  @AdrCreateSampleBufferFromSource := GetProcAddress(AdrDLL, '_AdrCreateSampleBufferFromSource@4');
  @AdrOpenSoundEffect              := GetProcAddress(AdrDLL, '_AdrOpenSoundEffect@12');
  @AdrOpenFile                     := GetProcAddress(AdrDLL, '_AdrOpenFile@8');
  @AdrCreateMemoryFile             := GetProcAddress(AdrDLL, '_AdrCreateMemoryFile@8');
  @AdrEnumerateCDDevices           := GetProcAddress(AdrDLL, '_AdrEnumerateCDDevices@0');
  @AdrOpenCDDevice                 := GetProcAddress(AdrDLL, '_AdrOpenCDDevice@4');
  @AdrOpenMIDIDevice               := GetProcAddress(AdrDLL, '_AdrOpenMIDIDevice@4');

  Result :=
    Assigned(AdrGetVersion) and
    Assigned(AdrGetSupportedFileFormats) and
    Assigned(AdrGetSupportedAudioDevices) and
    Assigned(AdrGetSampleSize) and
    Assigned(AdrOpenDevice) and
    Assigned(AdrOpenSampleSource) and
    Assigned(AdrOpenSampleSourceFromFile) and
    Assigned(AdrCreateTone) and
    Assigned(AdrCreateSquareWave) and
    Assigned(AdrCreateWhiteNoise) and
    Assigned(AdrCreatePinkNoise) and
    Assigned(AdrCreateLoopPointSource) and
    Assigned(AdrOpenSound) and
    Assigned(AdrCreateSampleBuffer) and
    Assigned(AdrCreateSampleBufferFromSource) and
    Assigned(AdrOpenSoundEffect) and
    Assigned(AdrOpenFile) and
    Assigned(AdrCreateMemoryFile) and
    Assigned(AdrEnumerateCDDevices) and
    Assigned(AdrOpenCDDevice) and
    Assigned(AdrOpenMIDIDevice);

  if not Result then
    AdrUnloadDLL
end;

procedure AdrUnloadDLL;
begin
  if AdrDLL = 0 then
    Exit;

  FreeLibrary(AdrDLL);
  AdrDLL := 0
end;

function AdrDLLLoaded: Boolean;
begin
  Result := (AdrDLL <> 0)
end;

{$ELSE}

(* static function links ******************************************************)

function AdrGetVersion: PChar; stdcall; external DLL_NAME name '_AdrGetVersion@0';
function AdrGetSupportedFileFormats: PChar; stdcall; external DLL_NAME name '_AdrGetSupportedFileFormats@0';
function AdrGetSupportedAudioDevices: PChar; stdcall; external DLL_NAME name '_AdrGetSupportedAudioDevices@0';
function AdrGetSampleSize(aFormat: TAdrSampleFormat): Integer; stdcall; external DLL_NAME name '_AdrGetSampleSize@4';
function AdrOpenDevice(const aName: PChar; const aParams: PChar): TAdrAudioDevice; stdcall; external DLL_NAME name '_AdrOpenDevice@8';
function AdrOpenSampleSource(const aFilename: PChar; aFileFormat: TAdrFileFormat): TAdrSampleSource; stdcall; external DLL_NAME name '_AdrOpenSampleSource@8';
function AdrOpenSampleSourceFromFile(aFile: TAdrFile; aFileFormat: TAdrFileFormat): TAdrSampleSource; stdcall; external DLL_NAME name '_AdrOpenSampleSourceFromFile@8';
function AdrCreateTone(aFrequency: Double): TAdrSampleSource; stdcall; external DLL_NAME name '_AdrCreateTone@8';
function AdrCreateSquareWave(aFrequency: Double): TAdrSampleSource; stdcall; external DLL_NAME name '_AdrCreateSquareWave@8';
function AdrCreateWhiteNoise: TAdrSampleSource; stdcall; external DLL_NAME name '_AdrCreateWhiteNoise@0';
function AdrCreatePinkNoise: TAdrSampleSource; stdcall; external DLL_NAME name '_AdrCreatePinkNoise@0';
function AdrCreateLoopPointSource(aSource: TAdrSampleSource): TAdrLoopPointSource; stdcall; external DLL_NAME name '_AdrCreateLoopPointSource@4';
function AdrOpenSound(aDevice: TAdrAudioDevice; aSource: TAdrSampleSource; aStreaming: LongBool): TAdrOutputStream; stdcall; external DLL_NAME name '_AdrOpenSound@12';
function AdrCreateSampleBuffer(aSamples: Pointer; aFrameCount, aChannelCount, aSampleRate: Integer; aSampleFormat: TAdrSampleFormat): TAdrSampleBuffer; stdcall; external DLL_NAME name '_AdrCreateSampleBuffer@20';
function AdrCreateSampleBufferFromSource(aSource: TAdrSampleSource): TAdrSampleBuffer; stdcall; external DLL_NAME name '_AdrCreateSampleBufferFromSource@4';
function AdrOpenSoundEffect(aDevice: TAdrAudioDevice; aSource: TAdrSampleSource; aType: TAdrSoundEffectType): TAdrSoundEffect; stdcall; external DLL_NAME name '_AdrOpenSoundEffect@12';
function AdrOpenFile(aName: PChar; aWriteable: Boolean): TAdrFile; stdcall; external DLL_NAME name '_AdrOpenFile@8';
function AdrCreateMemoryFile(aBuffer: Pointer; aSize: Integer): TAdrFile; stdcall; external DLL_NAME name '_AdrCreateMemoryFile@8';
function AdrEnumerateCDDevices: PChar; stdcall; external DLL_NAME name '_AdrEnumerateCDDevices@0';
function AdrOpenCDDevice(aName: PChar): TAdrCDDevice; stdcall; external DLL_NAME name '_AdrOpenCDDevice@4';
function AdrOpenMIDIDevice(aname: PChar): TAdrMIDIDevice; stdcall; external DLL_NAME name '_AdrOpenMIDIDevice@4';
{$ENDIF}

(* non-abstract method implementations ****************************************)

function TAdrStopEvent.GetType: TAdrEventType; stdcall;
begin
  Result := ET_STOP
end;

function TAdrStopCallback.GetType: TAdrEventType; stdcall;
begin
  Result := ET_STOP
end;

procedure TAdrStopCallback.Call(aEvent: TAdrEvent); stdcall;
begin
  StreamStopped(aEvent)
end;

(* public API functions *******************************************************)

function GetVersion: PChar;
begin
  Result := AdrGetVersion;
end;

procedure SplitString(var OutList: TAdrStringVect; InList: PChar; Delim: Char);
var
  Next: Integer;
begin
  SetLength(OutList, 0);

  while Length(InList) > 0 do
  begin
    SetLength(OutList, Length(OutList) + 1);

    Next := Pos(Delim, InList);
    if Next > 0 then
    begin
      OutList[High(OutList)] := Copy(InList, 1, Next - 1);
      InList := @InList[Next]
    end
    else
    begin
      OutList[High(OutList)] := Copy(InList, 1, Length(InList));
      Exit
    end
  end
end;

procedure GetSupportedFileFormats(var aFormats: TAdrFileFormatDesc);
var
  i,
  Colon       : Integer;
  Descriptions: TAdrStringVect;
begin
  SplitString(Descriptions, AdrGetSupportedFileFormats, ';');

  SetLength(aFormats, Length(Descriptions));

  for i := 0 to High(Descriptions) do
  begin
    Colon := Pos(':', Descriptions[i]);
    aFormats[i].Description := Copy(Descriptions[i], 1, Colon - 1);
    SplitString(aFormats[i].Extensions,
      PChar(Copy(Descriptions[i], Colon + 1, Length(Descriptions[i]) - Colon)),
      ',')
  end
end;

procedure GetSupportedAudioDevices(var aDevices: TAdrAudioDeviceDesc);
var
  i           : Integer;
  Description,
  Descriptions: TAdrStringVect;
begin
  SplitString(Descriptions, AdrGetSupportedAudioDevices, ';');

  SetLength(aDevices, Length(Descriptions));

  for i := 0 to High(Descriptions) do
  begin
    SplitString(Description, PChar(Descriptions[i]), ':');
    aDevices[i].Name := Description[0];
    aDevices[i].Description := Description[1]
  end
end;

function GetSampleSize(aFormat: TAdrSampleFormat): Integer;
begin
  Result := AdrGetSampleSize(aFormat)
end;

function OpenDevice(aName, aParameters: PChar): TAdrAudioDevice;
begin
  Result := AdrOpenDevice(aName, aParameters)
end;

function OpenSampleSource(aFilename: PChar; aFileFormat: TAdrFileFormat = FF_AUTODETECT): TAdrSampleSource; overload;
begin
  Result := AdrOpenSampleSource(aFilename, aFileFormat)
end;

function OpenSampleSource(aFile: TAdrFile; aFileFormat: TAdrFileFormat = FF_AUTODETECT): TAdrSampleSource; overload;
begin
  Result := AdrOpenSampleSourceFromFile(aFile, aFileFormat)
end;

function CreateTone(aFrequency: Double): TAdrSampleSource;
begin
  Result := AdrCreateTone(aFrequency)
end;

function CreateSquareWave(aFrequency: Double): TAdrSampleSource;
begin
  Result := AdrCreateSquareWave(aFrequency)
end;

function CreateWhiteNoise: TAdrSampleSource;
begin
  Result := AdrCreateWhiteNoise
end;

function CreatePinkNoise: TAdrSampleSource;
begin
  Result := AdrCreatePinkNoise
end;

function CreateLoopPointSource(aSource: TAdrSampleSource): TAdrLoopPointSource; overload;
begin
  Result := AdrCreateLoopPointSource(aSource)
end;

function CreateLoopPointSource(aFilename: PChar; aFileFormat: TAdrFileFormat = FF_AUTODETECT): TAdrLoopPointSource; overload;
begin
  Result := CreateLoopPointSource(OpenSampleSource(aFilename, aFileFormat))
end;

function CreateLoopPointSource(aFile: TAdrFile; aFileFormat: TAdrFileFormat = FF_AUTODETECT): TAdrLoopPointSource; overload;
begin
  Result := CreateLoopPointSource(OpenSampleSource(aFile, aFileFormat))
end;

function OpenSound(aDevice: TAdrAudioDevice; aSource: TAdrSampleSource; aStreaming: Boolean = False): TAdrOutputStream; overload;
begin
  Result := AdrOpenSound(aDevice, aSource, aStreaming)
end;

function OpenSound(aDevice: TAdrAudioDevice; aFilename: PChar; aStreaming: Boolean = False; aFileFormat: TAdrFileFormat = FF_AUTODETECT): TAdrOutputStream; overload;
var
  Source: TAdrSampleSource;
begin
  Source := OpenSampleSource(aFilename, aFileFormat);
  Result := OpenSound(aDevice, Source, aStreaming)
end;

function OpenSound(aDevice: TAdrAudioDevice; aFile: TAdrFile; aStreaming: Boolean = False; aFileFormat: TAdrFileFormat = FF_AUTODETECT): TAdrOutputStream; overload;
var
  Source: TAdrSampleSource;
begin
  Source := OpenSampleSource(aFile, aFileFormat);
  Result := OpenSound(aDevice, Source, aStreaming)
end;

function CreateSampleBuffer(aSamples: Pointer; aFrameCount, aChannelCount, aSampleRate: Integer; aSampleFormat: TAdrSampleFormat): TAdrSampleBuffer; overload;
begin
  Result := AdrCreateSampleBuffer(aSamples, aFrameCount, aChannelCount, aSampleRate, aSampleFormat)
end;

function CreateSampleBuffer(aSource: TAdrSampleSource): TAdrSampleBuffer; overload;
begin
  Result := AdrCreateSampleBufferFromSource(aSource)
end;

function OpenSoundEffect(aDevice: TAdrAudioDevice; aSource: TAdrSampleSource; aType: TAdrSoundEffectType): TAdrSoundEffect; overload;
begin
  Result := AdrOpenSoundEffect(aDevice, aSource, aType)
end;

function OpenSoundEffect(aDevice: TAdrAudioDevice; aFilename: PChar; aType: TAdrSoundEffectType; aFileFormat: TAdrFileFormat = FF_AUTODETECT): TAdrSoundEffect; overload;
var
  Source: TAdrSampleSource;
begin
  Source := OpenSampleSource(aFilename, aFileFormat);
  Result := OpenSoundEffect(aDevice, Source, aType)
end;

function OpenSoundEffect(aDevice: TAdrAudioDevice; aFile: TAdrFile; aType: TAdrSoundEffectType; aFileFormat: TAdrFileFormat = FF_AUTODETECT): TAdrSoundEffect; overload;
var
  Source: TAdrSampleSource;
begin
  Source := OpenSampleSource(aFile, aFileFormat);
  Result := OpenSoundEffect(aDevice, Source, aType)
end;

function OpenFile(aFilename: PChar; aWriteable: Boolean): TAdrFile;
begin
  Result := AdrOpenFile(aFilename, aWriteable)
end;

function CreateMemoryFile(aBuffer: Pointer; aSize: Integer): TAdrFile;
begin
  Result := AdrCreateMemoryFile(aBuffer, aSize)
end;

procedure EnumerateCDDevices(var aDevices: TAdrStringVect);
var
  DeviceList: PChar;
begin
  SetLength(aDevices, 0);

  DeviceList := AdrEnumerateCDDevices;

  while Length(DeviceList) > 0 do
  begin
    SetLength(aDevices, Length(aDevices) + 1);
    aDevices[High(aDevices)] := DeviceList;
    DeviceList := @DeviceList[Length(DeviceList) + 1]
  end
end;

function OpenCDDevice(aDevice: PChar): TAdrCDDevice;
begin
  Result := AdrOpenCDDevice(aDevice)
end;

function OpenMIDIDevice(aDevice: PChar): TAdrMIDIDevice;
begin
  Result := AdrOpenMIDIDevice(aDevice)
end;

initialization

finalization
{$IFDEF DYNAMIC}
  AdrUnloadDLL;
{$ENDIF}

end.