{ File   : Sound.pas
  Author : Deniz Oezmen
  Created: 2007-03-04
  Changed: 2010-08-29

  Simple Audiere wrapper.
}
unit Sound;

interface

function AssignSound(const SoundFileName: string): Boolean;
function IsSoundAssigned: Boolean;
procedure UnassignSound;
procedure PlaySound;
procedure PauseSound;
procedure StopSound;
function SoundIsPlaying: Boolean;
function GetSoundLength: Integer;
procedure SetSoundPosition(Position: Integer);
function GetSoundPosition: Integer;

implementation

uses
  SysUtils, Audiere;

var
  AudioDevice: TAdrAudioDevice;
  SoundStream: TAdrOutputStream;

function AssignSound(const SoundFileName: string): Boolean;
begin
  Result := False;

  if not Assigned(AudioDevice) then
    Exit;

  UnassignSound;
  SoundStream := OpenSound(AudioDevice, PChar(SoundFileName), True);

  Result := Assigned(SoundStream);
  if Result then
    SoundStream.Ref
end;

function IsSoundAssigned: Boolean;
begin
  Result := Assigned(SoundStream)
end;

procedure UnassignSound;
begin
  if not Assigned(SoundStream) then
    Exit;

  if SoundStream.IsPlaying then
    SoundStream.Stop;

  SoundStream.Unref;
  SoundStream := nil
end;

procedure PlaySound;
begin
  if not Assigned(SoundStream) then
    Exit;

  SoundStream.Play
end;

procedure PauseSound;
begin
  if not Assigned(SoundStream) then
    Exit;

  SoundStream.Stop
end;

procedure StopSound;
begin
  if not Assigned(SoundStream) then
    Exit;

  SoundStream.Stop;
  SoundStream.Reset
end;

function SoundIsPlaying: Boolean;
begin
  Result := False;

  if not Assigned(SoundStream) then
    Exit;

  Result := SoundStream.IsPlaying
end;

function GetSoundLength: Integer;
begin
  Result := 0;

  if not Assigned(SoundStream) then
    Exit;

  Result := SoundStream.GetLength
end;

procedure SetSoundPosition(Position: Integer);
begin
  if not Assigned(SoundStream) then
    Exit;

  SoundStream.SetPosition(Position)
end;

function GetSoundPosition: Integer;
begin
  Result := 0;

  if not Assigned(SoundStream) then
    Exit;

  Result := SoundStream.GetPosition
end;

initialization
  AudioDevice := nil;

  // prepare sound system
  if AdrLoadDLL(ExtractFilePath(ParamStr(0))) then
    // open default audio device
    AudioDevice := OpenDevice('', '');

  if Assigned(AudioDevice) then
    AudioDevice.Ref;

  SoundStream := nil;

finalization
  UnassignSound;

  if Assigned(AudioDevice) then
  begin
    AudioDevice.Unref;
    AudioDevice := nil
  end

end.
 