{ File   : ANI.pas
  Author : Deniz Oezmen
  Created: 2007-03-14
  Changed: 2009-04-12

  The Longest Journey ANI reader.
}
unit ANI;

interface

uses
  Classes, Animation;

type
  TANICallback = procedure(const Frac: Double);

function LoadANI(const ANIName: string; out IntAnimation: TIntAnimation;
  const ANICallback: TANICallback): Boolean;
function LoadANIAsText(const ANIName: string; var ResourceText: TStringList;
  const ANICallback: TANICallback): Boolean;

implementation

uses
  TLJ_FS,
  SysUtils, StrUtils;

var
  ANIFile: TFS_Handle;

function ReadHeader(out IntAnimation: TIntAnimation): Boolean;
var
  BytesRead: Longword;
begin
  Result := False;

  with IntAnimation.Header do
  begin
    BytesRead := FS_BlockRead(ANIFile, ID, SizeOf(ID));
    if (BytesRead <> SizeOf(ID)) or (ID <> 3) then
      Exit;

    BytesRead := FS_BlockRead(ANIFile, Version, SizeOf(Version));
    if (BytesRead <> SizeOf(Version)) or
      ((Version <> 3) and (Version <> 256)) then
      Exit;

    if Version = 3 then
    begin
      Unknown1 := 0;
      FS_BlockRead(ANIFile, MaxTime, SizeOf(MaxTime));
      FS_BlockRead(ANIFile, Unknown2, SizeOf(Unknown2))
    end
    else
    begin
      FS_BlockRead(ANIFile, Unknown1, SizeOf(Unknown1));
      FS_BlockRead(ANIFile, Unknown2, SizeOf(Unknown2));
      FS_BlockRead(ANIFile, MaxTime, SizeOf(MaxTime))
    end
  end;

  Result := True
end;

function ReadSingle: Single;
begin
  FS_BlockRead(ANIFile, Result, SizeOf(Result))
end;

function ReadLongint: Longint;
begin
  FS_BlockRead(ANIFile, Result, SizeOf(Result))
end;

procedure ReadANI(var IntAnimation: TIntAnimation;
  const ANICallback: TANICallback);
var
  i, j: Longint;
begin
  with IntAnimation do
  begin
    SetLength(BoneAnims, ReadLongint);

    for i := 0 to High(BoneAnims) do
    begin
      with BoneAnims[i] do
      begin
        BoneIndex := ReadLongint;
        SetLength(KeyDatas, ReadLongint);

        for j := 0 to High(KeyDatas) do
          with KeyDatas[j] do
          begin
            AnimTime := ReadLongint;
            QRotX := ReadSingle;
            QRotY := ReadSingle;
            QRotZ := ReadSingle;
            QRotW := ReadSingle;
            PosX := ReadSingle;
            PosY := ReadSingle;
            PosZ := ReadSingle
          end
      end;

      ANICallback((i + 1) / Length(BoneAnims))
    end
  end
end;

function LoadANI(const ANIName: string; out IntAnimation: TIntAnimation;
  const ANICallback: TANICallback): Boolean;
begin
  Result := FS_OpenFile(ANIName, ANIFile);
  if not Result then
    Exit;

  Result := ReadHeader(IntAnimation);
  if not Result then
    Exit;

  ReadANI(IntAnimation, ANICallback);

  FS_CloseFile(ANIFile)
end;

function LoadANIAsText(const ANIName: string; var ResourceText: TStringList;
  const ANICallback: TANICallback): Boolean;
var
  i, j        : Longint;
  IntAnimation: TIntAnimation;
  OldSep      : Char;
begin
  Result := LoadANI(ANIName, IntAnimation, ANICallback);
  if not Result then
    Exit;

  OldSep := DecimalSeparator;
  DecimalSeparator := '.';

  ResourceText.Add('{');

  with IntAnimation do
  begin
    with Header do
    begin
      ResourceText.Add(Format('  Unknown.H.1 = %d,', [Unknown1]));
      ResourceText.Add(Format('  Unknown.H.2 = 0x%.8x,', [Unknown2]));
      ResourceText.Add(Format('  MaxTime = %d,', [MaxTime]))
    end;

    ResourceText.Add('  BoneAnims = [');
    for i := 0 to High(BoneAnims) do
      with BoneAnims[i] do
      begin
        ResourceText.Add(Format('    BoneAnim #%d = {', [i]));
        ResourceText.Add(Format('      BoneIndex = %d,', [BoneIndex]));

        ResourceText.Add('      KeyDatas = [');
        for j := 0 to High(KeyDatas) do
          with KeyDatas[j] do
          begin
            ResourceText.Add(Format('        KeyData #%d = {', [j]));
            ResourceText.Add(Format('          AnimTime = %d,', [AnimTime]));
            ResourceText.Add(Format('          QRotX = %1.7f,', [QRotX]));
            ResourceText.Add(Format('          QRotY = %1.7f,', [QRotY]));
            ResourceText.Add(Format('          QRotZ = %1.7f,', [QRotZ]));
            ResourceText.Add(Format('          QRotW = %1.7f,', [QRotW]));
            ResourceText.Add(Format('          PosX = %1.7f,', [PosX]));
            ResourceText.Add(Format('          PosY = %1.7f,', [PosY]));
            ResourceText.Add(Format('          PosZ = %1.7f',  [PosZ]));

            if j < High(KeyDatas) then
              ResourceText.Add('        },')
            else
              ResourceText.Add('        }')
          end;

        ResourceText.Add('      ]');

        if i < High(BoneAnims) then
          ResourceText.Add('    },')
        else
          ResourceText.Add('    }');

        ANICallback((i + 1) / Length(BoneAnims))
      end;

    ResourceText.Add('  ]')
  end;

  ResourceText.Add('}');

  DecimalSeparator := OldSep
end;

end.


