{ File   : CIR.pas
  Author : Deniz Oezmen
  Created: 2007-03-13, based on CIRRead.dpr created 2007-03-12
  Changed: 2009-04-12

  The Longest Journey CIR reader.
}
unit CIR;

interface

uses
  Classes, Animation;

type
  TCIRCallback = procedure(const Frac: Double);

function LoadCIR(const CIRName: string; out Model: TModel;
  const CIRCallback: TCIRCallback): Boolean;
function LoadCIRAsText(const CIRName: string; var ResourceText: TStringList;
  const CIRCallback: TCIRCallback): Boolean;

implementation

uses
  TLJ_FS,
  SysUtils, StrUtils;

type
  TCIRHeader = packed record
                 ID,                // 4
                 Version,           // 16 or 256
                 Unknown1,          // only present in version 256
                 Unknown2: Longint; // $deadbabe
                 Unknown3: Single
               end;

var
  CIRFile: TFS_Handle;

function ReadHeader(out Model: TModel): Boolean;
var
  BytesRead: Longword;
begin
  Result := False;

  with Model.Header do
  begin
    BytesRead := FS_BlockRead(CIRFile, ID, SizeOf(ID));
    if (BytesRead <> SizeOf(ID)) or (ID <> 4) then
      Exit;

    BytesRead := FS_BlockRead(CIRFile, Version, SizeOf(Version));
    if (BytesRead <> SizeOf(Version)) or
      ((Version <> 16) and (Version <> 256)) then
      Exit;

    if Version = 16 then
      Unknown1 := 0
    else
      FS_BlockRead(CIRFile, Unknown1, SizeOf(Unknown1));

    FS_BlockRead(CIRFile, Unknown2, SizeOf(Unknown2));
    FS_BlockRead(CIRFile, Unknown3, SizeOf(Unknown3))
  end;

  Result := True
end;

function ReadString: string;
var
  Len: Longword;
begin
  FS_BlockRead(CIRFile, Len, SizeOf(Len));
  SetLength(Result, Len);
  if Len > 0 then
    FS_BlockRead(CIRFile, Result[1], Len)
end;

function ReadSingle: Single;
begin
  FS_BlockRead(CIRFile, Result, SizeOf(Result))
end;

function ReadLongint: Longint;
begin
  FS_BlockRead(CIRFile, Result, SizeOf(Result))
end;

procedure ReadCIR(var Model: TModel; const CIRCallback: TCIRCallback);
var
  i, j, k: Longint;
begin
  with Model do
  begin
    SetLength(Materials, ReadLongint);
    for i := 0 to High(Materials) do
      with Materials[i] do
      begin
        Name := ReadString;
        Unknown1 := ReadLongint;
        TextureName := ReadString;
        ColourR := ReadSingle;
        ColourG := ReadSingle;
        ColourB := ReadSingle
      end;

    CIRCallback(0.25);

    SetLength(Unknown4, ReadLongint);
    for i := 0 to High(Unknown4) do
      with Unknown4[i] do
      begin
        Unknown4_1 := ReadSingle;
        Unknown4_2 := ReadSingle;
        Unknown4_3 := ReadSingle;
        Unknown4_4 := ReadSingle
      end;

    CIRCallback(0.5);

    SetLength(Skeleton, ReadLongint);
    for i := 0 to High(Skeleton) do
      with Skeleton[i] do
      begin
        Name := ReadString;
        Unknown1 := ReadSingle;

        SetLength(Children, ReadLongint);
        for j := 0 to High(Children) do
          Children[j] := ReadLongint
      end;


    CIRCallback(0.75);

    SetLength(Groups, ReadLongint);
    for i := 0 to High(Groups) do
      with Groups[i] do
      begin
        Name := ReadString;

        SetLength(Faces, ReadLongint);
        for j := 0 to High(Faces) do
          with Faces[j] do
          begin
            MaterialIndex := ReadLongint;

            SetLength(Vertices, ReadLongint);
            for k := 0 to High(Vertices) do
              with Vertices[k] do
              begin
                PosX1 := ReadSingle;
                PosY1 := ReadSingle;
                PosZ1 := ReadSingle;
                PosX2 := ReadSingle;
                PosY2 := ReadSingle;
                PosZ2 := ReadSingle;
                NormalX := ReadSingle;
                NormalY := ReadSingle;
                NormalZ := ReadSingle;
                TextureS := ReadSingle;
                TextureT := ReadSingle;
                BoneIndex1 := ReadLongint;
                BoneIndex2 := ReadLongint;
                BoneWeight := ReadSingle
              end;

            SetLength(Triangles, ReadLongint);
            for k := 0 to High(Triangles) do
              with Triangles[k] do
              begin
                VertexIndex1 := ReadLongint;
                VertexIndex2 := ReadLongint;
                VertexIndex3 := ReadLongint
              end
          end;

        SetLength(Unknown1, ReadLongint);
        for j := 0 to High(Unknown1) do
          with Unknown1[j] do
          begin
            Unknown1_1 := ReadSingle;
            Unknown1_2 := ReadSingle;
            Unknown1_3 := ReadSingle;
            Unknown1_4 := ReadSingle;
            Unknown1_5 := ReadLongint
          end;

        SetLength(Unknown2, ReadLongint);
        for j := 0 to High(Unknown2) do
          with Unknown2[j] do
          begin
            Unknown2_01 := ReadString;
            Unknown2_02 := ReadSingle;
            Unknown2_03 := ReadSingle;
            Unknown2_04 := ReadSingle;
            Unknown2_05 := ReadSingle;
            Unknown2_06 := ReadSingle;
            Unknown2_07 := ReadSingle;
            Unknown2_08 := ReadSingle;
            Unknown2_09 := ReadSingle;
            Unknown2_10 := ReadLongint
          end
      end
  end;

  CIRCallback(1)
end;

function LoadCIR(const CIRName: string; out Model: TModel;
  const CIRCallback: TCIRCallback): Boolean;
begin
  Result := FS_OpenFile(CIRName, CIRFile);
  if not Result then
    Exit;

  Result := ReadHeader(Model);
  if not Result then
    Exit;

  ReadCIR(Model, CIRCallback);

  FS_CloseFile(CIRFile)
end;

function LoadCIRAsText(const CIRName: string; var ResourceText: TStringList;
  const CIRCallback: TCIRCallback): Boolean;
var
  i, j, k: Longint;
  Model  : TModel;
  OldSep : Char;
begin
  Result := LoadCIR(CIRName, Model, CIRCallback);
  if not Result then
    Exit;

  OldSep := DecimalSeparator;
  DecimalSeparator := '.';

  ResourceText.Add('{');

  with Model do
  begin
    with Header do
    begin
      ResourceText.Add(Format('  Unknown.H.1 = %d,', [Unknown1]));
      ResourceText.Add(Format('  Unknown.H.2 = 0x%.8x,', [Unknown2]));
      ResourceText.Add(Format('  Unknown.H.3 = %1.7f,', [Unknown3]))
    end;

    ResourceText.Add('  Materials = [');
    for i := 0 to High(Materials) do
      with Materials[i] do
      begin
        ResourceText.Add(Format('    Material #%d = {', [i]));
        ResourceText.Add(Format('      Name = "%s",', [Name]));
        ResourceText.Add(Format('      Unknown.M.1 = %d,', [Unknown1]));
        ResourceText.Add(Format('      TextureName = "%s",', [TextureName]));
        ResourceText.Add(Format('      ColourR = %1.7f,', [ColourR]));
        ResourceText.Add(Format('      ColourG = %1.7f,', [ColourG]));
        ResourceText.Add(Format('      ColourB = %1.7f', [ColourB]));

        if i < High(Materials) then
          ResourceText.Add('    },')
        else
          ResourceText.Add('    }')
      end;

    ResourceText.Add('  ],');

    CIRCallback(0.25);

    ResourceText.Add('  UnknownList.1 = [');
    for i := 0 to High(Unknown4) do
      with Unknown4[i] do
      begin
        ResourceText.Add(Format('    Unknown.1.1 #%d = {', [i]));
        ResourceText.Add(Format('      Unknown.1.1.1 = %1.7f,', [Unknown4_1]));
        ResourceText.Add(Format('      Unknown.1.1.2 = %1.7f,', [Unknown4_2]));
        ResourceText.Add(Format('      Unknown.1.1.3 = %1.7f,', [Unknown4_3]));
        ResourceText.Add(Format('      Unknown.1.1.4 = %1.7f', [Unknown4_4]));

        if i < High(Unknown4) then
          ResourceText.Add('    },')
        else
          ResourceText.Add('    }')
      end;

    ResourceText.Add('  ],');

    CIRCallback(0.5);

    ResourceText.Add('  Skeleton = [');
    for i := 0 to High(Skeleton) do
      with Skeleton[i] do
      begin
        ResourceText.Add(Format('    Bone #%d = {', [i]));
        ResourceText.Add(Format('      Name = "%s",', [Name]));
        ResourceText.Add(Format('      Unknown.2.1.2 = %1.7f,', [Unknown1]));

        ResourceText.Add('      Children = [');
        for j := 0 to High(Children) do
        begin
          if j < High(Children) then
            ResourceText.Add(Format('        Child #%d = %d,',
              [j, Children[j]]))
          else
            ResourceText.Add(Format('        Child #%d = %d',
              [j, Children[j]]))
        end;
        ResourceText.Add('      ]');

        if i < High(Skeleton) then
          ResourceText.Add('    },')
        else
          ResourceText.Add('    }')
      end;

    ResourceText.Add('  ],');

    CIRCallback(0.75);

    ResourceText.Add('  Groups  = [');
    for i := 0 to High(Groups) do
      with Groups[i] do
      begin
        ResourceText.Add(Format('    Group #%d = {', [i]));
        ResourceText.Add(Format('      Name = "%s",', [Name]));

        ResourceText.Add('      Faces = [');
        for j := 0 to High(Faces) do
          with Faces[j] do
          begin
            ResourceText.Add(Format('        Face1 #%d = {', [j]));
            ResourceText.Add(Format('          MaterialIndex = %d,',
              [MaterialIndex]));

            ResourceText.Add('          Vertices = [');
            for k := 0 to High(Vertices) do
              with Vertices[k] do
              begin
                ResourceText.Add(Format('            Vertex #%d = {',
                  [k]));
                ResourceText.Add(Format('              PosX1 = %1.7f,',
                  [PosX1]));
                ResourceText.Add(Format('              PosY1 = %1.7f,',
                  [PosY1]));
                ResourceText.Add(Format('              PosZ1 = %1.7f,',
                  [PosZ1]));
                ResourceText.Add(Format('              PosX2 = %1.7f,',
                  [PosX2]));
                ResourceText.Add(Format('              PosY2 = %1.7f,',
                  [PosY2]));
                ResourceText.Add(Format('              PosZ2 = %1.7f,',
                  [PosZ2]));
                ResourceText.Add(Format('              NormalX = %1.7f,',
                  [NormalX]));
                ResourceText.Add(Format('              NormalY = %1.7f,',
                  [NormalY]));
                ResourceText.Add(Format('              NormalZ = %1.7f,',
                  [NormalZ]));
                ResourceText.Add(Format('              TextureS = %1.7f,',
                  [TextureS]));
                ResourceText.Add(Format('              TextureT = %1.7f,',
                  [TextureT]));
                ResourceText.Add(Format('              BoneIndex1 = %d,',
                  [BoneIndex1]));
                ResourceText.Add(Format('              BoneIndex2 = %d,',
                  [BoneIndex2]));
                ResourceText.Add(Format('              BoneWeight = %1.7f',
                  [BoneWeight]));

                if k < High(Vertices) then
                  ResourceText.Add('            },')
                else
                  ResourceText.Add('            }')
              end;

            ResourceText.Add('          ],');

            ResourceText.Add('          Triangles = [');
            for k := 0 to High(Triangles) do
              with Triangles[k] do
              begin
                ResourceText.Add(Format('            Triangle #%d = {',
                  [k]));
                ResourceText.Add(Format('              VertexIndex1 = %d,',
                  [VertexIndex1]));
                ResourceText.Add(Format('              VertexIndex2 = %d,',
                  [VertexIndex2]));
                ResourceText.Add(Format('              VertexIndex3 = %d',
                  [VertexIndex3]));

                if k < High(Triangles) then
                  ResourceText.Add('            },')
                else
                  ResourceText.Add('            }')
              end;

            ResourceText.Add('          ]');

            if j < High(Faces) then
              ResourceText.Add('        },')
            else
              ResourceText.Add('        }')
          end;

        ResourceText.Add('      ],');

        ResourceText.Add('      UnknownList.3.1.3 = [');
        for j := 0 to High(Unknown1) do
          with Unknown1[j] do
          begin
            ResourceText.Add(Format('        Unknown.3.1.3.1 #%d = {', [j]));
            ResourceText.Add(Format('          Unknown.3.1.3.1.1 = %1.7f,',
              [Unknown1_1]));
            ResourceText.Add(Format('          Unknown.3.1.3.1.2 = %1.7f,',
              [Unknown1_2]));
            ResourceText.Add(Format('          Unknown.3.1.3.1.3 = %1.7f,',
              [Unknown1_3]));
            ResourceText.Add(Format('          Unknown.3.1.3.1.4 = %1.7f,',
              [Unknown1_4]));
            ResourceText.Add(Format('          Unknown.3.1.3.1.5 = %d',
              [Unknown1_5]));

            if j < High(Unknown1) then
              ResourceText.Add('        },')
            else
              ResourceText.Add('        }')
          end;

        ResourceText.Add('      ],');

        ResourceText.Add('      UnknownList.3.1.4 = [');
        for j := 0 to High(Unknown2) do
          with Unknown2[j] do
          begin
            ResourceText.Add(Format('        Unknown.3.1.4.1 #%d = {', [j]));
            ResourceText.Add(Format('          Unknown.3.1.4.1.1 = "%s",',
              [Unknown2_01]));
            ResourceText.Add(Format('          Unknown.3.1.4.1.2 = %1.7f,',
              [Unknown2_02]));
            ResourceText.Add(Format('          Unknown.3.1.4.1.3 = %1.7f,',
              [Unknown2_03]));
            ResourceText.Add(Format('          Unknown.3.1.4.1.4 = %1.7f,',
              [Unknown2_04]));
            ResourceText.Add(Format('          Unknown.3.1.4.1.5 = %1.7f,',
              [Unknown2_05]));
            ResourceText.Add(Format('          Unknown.3.1.4.1.6 = %1.7f,',
              [Unknown2_06]));
            ResourceText.Add(Format('          Unknown.3.1.4.1.7 = %1.7f,',
              [Unknown2_07]));
            ResourceText.Add(Format('          Unknown.3.1.4.1.8 = %1.7f,',
              [Unknown2_08]));
            ResourceText.Add(Format('          Unknown.3.1.4.1.9 = %1.7f,',
              [Unknown2_09]));
            ResourceText.Add(Format('          Unknown.3.1.4.1.10 = %d',
              [Unknown2_10]));

            if j < High(Unknown2) then
              ResourceText.Add('        },')
            else
              ResourceText.Add('        }')
          end;

        ResourceText.Add('      ]');

        if i < High(Groups) then
          ResourceText.Add('    },')
        else
          ResourceText.Add('    }')
      end;

    ResourceText.Add('  ]')
  end;

  ResourceText.Add('}');

  DecimalSeparator := OldSep;

  CIRCallback(1)
end;

end.


