{ File   : Raw.pas
  Author : Deniz Oezmen
  Created: 2006-11-04
  Changed: 2009-04-12

  The Longest Journey general raw data loader.
}
unit Raw;

interface

uses
  Classes;

type
  TRawBuffer   = array of Byte;

  TRawCallback = procedure(const Frac: Double);

function LoadRaw(const RawName: string; out RawBuffer: TRawBuffer;
  const RawCallback: TRawCallback): Boolean;
function LoadRawAsHex(const RawName: string; var HexData: TStringList;
  const RawCallback: TRawCallback): Boolean;

implementation

uses
  TLJ_FS,
  SysUtils;

function LoadRaw(const RawName: string; out RawBuffer: TRawBuffer;
  const RawCallback: TRawCallback): Boolean;
const
  BufSize = 1048576;
var
  RawFile: TFS_Handle;
begin
  Result := FS_OpenFile(RawName, RawFile);
  if not Result then
    Exit;

  SetLength(RawBuffer, FS_FileSize(RawFile));

  while not FS_Eof(RawFile) do
  begin
    FS_BlockRead(RawFile, RawBuffer[FS_FilePos(RawFile)], BufSize);
    RawCallback(FS_FilePos(RawFile) / FS_FileSize(RawFile))
  end;

  FS_CloseFile(RawFile)
end;

function ToHexViewChar(b: Byte): Char;
begin
  if b in [33..127] then
    Result := Chr(b)
  else
    Result := '.'
end;

function LoadRawAsHex(const RawName: string; var HexData: TStringList;
  const RawCallback: TRawCallback): Boolean;
const
  MaxDataSize = 1 shl 17;
var
  i, j, k  : Longint;
  Line     : string;
  RawBuffer: TRawBuffer;
begin
  Result := LoadRaw(RawName, RawBuffer, RawCallback);
  if not Result then
    Exit;

  if Length(RawBuffer) > MaxDataSize then
    SetLength(RawBuffer, MaxDataSize);

  for i := 0 to (Length(RawBuffer) shr 4) - 1 do
  begin
    Line := Format('%.4x:%.4x | ', [i shr 12, (i shl 4) and $ffff]);
    j := i shl 4;

    for k := 0 to 15 do
      Line := Format('%s%.2x ', [Line, RawBuffer[j + k]]);

    Line := Line + '| ';

    for k := 0 to 15 do
      Line := Format('%s%s', [Line, ToHexViewChar(RawBuffer[j + k])]);
    HexData.Add(Line);

    RawCallback((i + 1) / (Length(RawBuffer) shr 4))
  end;

  if Length(RawBuffer) and $0f <> 0 then
  begin
    i := Length(RawBuffer);
    Line := Format('%.4x:%.4x | ', [i shr 16, i and $ffff]);

    for i := (Length(RawBuffer) shr 4) shl 4 to Length(RawBuffer) - 1 do
      Line := Format('%s%.2x ', [Line, RawBuffer[i]]);

    while Length(Line) < 60 do
      Line := Line + ' ';

    Line := Line + '| ';

    for i := (Length(RawBuffer) shr 4) shl 4 to Length(RawBuffer) - 1 do
      Line := Format('%s%s', [Line, ToHexViewChar(RawBuffer[i])]);
    HexData.Add(Line);
  end;

  Finalize(RawBuffer)
end;

end.
