{ File    : $URL: file:///D:/SVN/svnDelphi/trunk/TLJ/TLJView/src/convert/XRC.pas $
  Author  : $Author: Deniz Oezmen $
  Changed : $Date: 2016-01-01 23:54:37 +0100 (Fr, 01 Jan 2016) $
  Revision: $Rev: 410 $

  The Longest Journey XRC reader.
}
unit XRC;

interface

uses
  Classes;

type
  TXRCCallback = procedure(const Frac: Double);

function ConvertXRC(const XRCName: string; var ResourceText: TStringList;
  const XRCCallback: TXRCCallback): Boolean;

implementation

uses
  TLJ_FS,
  SysUtils;

type
  TXRCEntry = record
                TypeID,
                Tag1       : Byte;
                Tag2,
                NameLen    : Word;
                Name       : string;
                DataSize   : Longint;
                NumChildren,
                Tag4       : Word
              end;

var
  Buffer : array of Byte;
  XRCFile: TFS_Handle;

procedure ReadXRC(RecLevel: Longword; var ResourceText: TStringList); forward;

procedure WriteAnimScriptLineDetails(RecLevel: Longword; Tag1: Byte;
  var ResourceText: TStringList);
var
  Ofs,
  Code,
  LOp1,
  LOp2: Longint;
  WOp1,
  WOp2: Smallint;
  Line: string;
begin
  Ofs := 0;
  Move(Buffer[Ofs], Code, SizeOf(Code));
  Inc(Ofs, SizeOf(Code));

  case Code of
    0:
      begin
        Move(Buffer[Ofs], LOp1, SizeOf(LOp1));
        Inc(Ofs, SizeOf(LOp1));
        Move(Buffer[Ofs], LOp2, SizeOf(LOp2));
        Inc(Ofs, SizeOf(LOp2));
        Line := Format('opcode: SHOW; pause: %d ms; frame: %d', [LOp1, LOp2])
      end;
    2:
      begin
        Move(Buffer[Ofs], LOp1, SizeOf(LOp1));
        Inc(Ofs, SizeOf(LOp1));
        Move(Buffer[Ofs], LOp2, SizeOf(LOp2));
        Inc(Ofs, SizeOf(LOp2));
        Line := Format('opcode: GOTO; pause: %d ms; line: %d', [LOp1, LOp2])
      end;
    3:
      begin
        Move(Buffer[Ofs], LOp1, SizeOf(LOp1));
        Inc(Ofs, SizeOf(LOp1));
        Move(Buffer[Ofs], WOp1, SizeOf(WOp1));
        Inc(Ofs, SizeOf(WOp1));
        Move(Buffer[Ofs], WOp2, SizeOf(WOp2));
        Inc(Ofs, SizeOf(WOp2));
        Line := Format('opcode: SHOW_RANDOM_FRAME; pause: %d ms; max. frame: ' +
          '%d; min. frame: %d', [LOp1, WOp1, WOp2])
      end;
    4:
      begin
        Move(Buffer[Ofs], LOp1, SizeOf(LOp1));
        Inc(Ofs, SizeOf(LOp1));
        Move(Buffer[Ofs], LOp2, SizeOf(LOp2));
        Inc(Ofs, SizeOf(LOp2));
        Line := Format('opcode: WAIT_RANDOM_TIME; pause: %d ms; max. wait: %d' +
          ' ms', [LOp1, LOp2])
      end
  else
    Line := 'opcode: unknown'
  end;

  ResourceText.Add(Format('%-*s[%s]', [RecLevel + 2, '', Line]));
  
  if Ofs < Length(Buffer) then
    ResourceText.Add(Format('%-*s[%d bytes of unknown data]', [RecLevel + 2, '',
      Length(Buffer) - Ofs]))
end;

procedure WriteObjectDetails(RecLevel: Longword; Tag1: Byte;
  var ResourceText: TStringList);
var
  XPos,
  YPos : Longint;
begin
  if not (Tag1 in [5, 6, 8]) then
  begin
    ResourceText.Add(Format('%-*s[%d bytes of unknown data]', [RecLevel + 2, '',
      Length(Buffer)]));
    Exit
  end;

  ResourceText.Add(Format('%-*s[16 bytes of unknown data]', [RecLevel + 2,
    '']));

  Move(Buffer[16], XPos, 4);
  Move(Buffer[20], YPos, 4);

  ResourceText.Add(Format('%-*s[XPos: %d, YPos: %d]', [RecLevel + 2, '', XPos,
    YPos]));

  if Length(Buffer) > 24 then
    ResourceText.Add(Format('%-*s[%d bytes of unknown data]', [RecLevel + 2,
      '', Length(Buffer) - 24]))
end;

procedure WriteAnimationDetails(RecLevel: Longword; Tag1: Byte;
  var ResourceText: TStringList);
var
  i,
  Ofs     : Longint;
  Size    : Word;
  CatStr,
  Line    : string;
  AnimDesc: packed record
              Category : Longint;
              NumFrames: Longint;
            end;
  SprAnim : packed record
              Width    : Longint;
              Height   : Longint;
              NumFrames: Longint;
            end;
  FrameDsc: packed record
              XPos    : Longint;
              YPos    : Longint;
              Unknown1: Longint;
              Unknown2: Longint;
              Width   : Longint;
              Height  : Longint;
            end;
begin
  // 3: sprite animation, 4: 3D animation?
  if not (Tag1 in [3, 4]) then
  begin
    ResourceText.Add(Format('%-*s[%d bytes of unknown data]', [RecLevel + 2, '',
      Length(Buffer)]));
    Exit
  end;

  Ofs := 0;
  Move(Buffer[Ofs], AnimDesc, SizeOf(AnimDesc));
  Inc(Ofs, SizeOf(AnimDesc));
  Move(Buffer[Ofs], Size, SizeOf(Size));
  Inc(Ofs, SizeOf(Size));
  SetLength(Line, Size);
  if Size > 0 then
    Move(Buffer[Ofs], Line[1], Size);
  Inc(Ofs, Size);

  case AnimDesc.Category of
    0: CatStr := 'unspecified';
    1: CatStr := 'idle';
    2: CatStr := 'walk';
    3: CatStr := 'talk'
  else
    CatStr := 'unknown'
  end;

  ResourceText.Add(Format('%-*s[category: %s, number of frames: %d, animation' +
    ' file: "%s"]', [RecLevel + 2, '', CatStr, AnimDesc.NumFrames, Line]));

  if Tag1 = 3 then
  begin
    Move(Buffer[Ofs], SprAnim, SizeOf(SprAnim));
    Inc(Ofs, SizeOf(SprAnim));

    with SprAnim do
      ResourceText.Add(Format('%-*s[width: %d, height: %d, number of frames: ' +
        '%d]', [RecLevel + 2, '', Width, Height, NumFrames]));

    for i := 1 to AnimDesc.NumFrames do
    begin
      Move(Buffer[Ofs], FrameDsc, SizeOf(FrameDsc));
      Inc(Ofs, SizeOf(FrameDsc));

      with FrameDsc do
        ResourceText.Add(Format('%-*s[XPos: %d, YPos: %d, unknown1: %d, unkno' +
          'wn2: %d, width: %d, height: %d]', [RecLevel + 2, '', XPos, YPos,
          Unknown1, Unknown2, Width, Height]))
    end
  end;

  if Ofs < Length(Buffer) then
    ResourceText.Add(Format('%-*s[%d bytes of unknown data]', [RecLevel + 2, '',
      Length(Buffer) - Ofs]))
end;

procedure WriteImageDetails(RecLevel: Longword; Tag1: Byte;
  var ResourceText: TStringList);
var
  Ofs : Longint;
  Size: Word;
  Line: string;
begin
  Ofs := 0;
  Move(Buffer[Ofs], Size, SizeOf(Size));
  Inc(Ofs, SizeOf(Size));
  SetLength(Line, Size);
  if Size > 0 then
    Move(Buffer[Ofs], Line[1], Size);
  Inc(Ofs, Size);

  ResourceText.Add(Format('%-*s[image file: "%s"]', [RecLevel + 2, '', Line]));

  if Ofs < Length(Buffer) then
    ResourceText.Add(Format('%-*s[%d bytes of unknown data]', [RecLevel + 2, '',
      Length(Buffer) - Ofs]))
end;

procedure WriteSoundDetails(RecLevel: Longword; Tag1: Byte;
  var ResourceText: TStringList);
var
  Ofs : Longint;
  Size: Word;
  Line: string;
begin
  Ofs := 0;
  Move(Buffer[Ofs], Size, SizeOf(Size));
  Inc(Ofs, SizeOf(Size));
  SetLength(Line, Size);
  if Size > 0 then
    Move(Buffer[Ofs], Line[1], Size);
  Inc(Ofs, Size);

  ResourceText.Add(Format('%-*s[sound file: "%s"]', [RecLevel + 2, '', Line]));

  Inc(Ofs, 28);

  ResourceText.Add(Format('%-*s[28 bytes of unknown data]', [RecLevel + 2,
    '']));

  Move(Buffer[Ofs], Size, SizeOf(Size));
  Inc(Ofs, SizeOf(Size));
  SetLength(Line, Size);
  if Size > 0 then
    Move(Buffer[Ofs], Line[1], Size);
  Inc(Ofs, Size);

  ResourceText.Add(Format('%-*s[local ID: "%s"]', [RecLevel + 2, '', Line]));

  if Ofs < Length(Buffer) then
    ResourceText.Add(Format('%-*s[%d bytes of unknown data]', [RecLevel + 2, '',
      Length(Buffer) - Ofs]))
end;

procedure WriteScriptVarDetails(RecLevel: Longword; Tag1: Byte;
  var ResourceText: TStringList);
var
  Ofs      : Longint;
  ScriptVar: Longword;
begin
  Ofs := 0;

  if Ofs + SizeOf(ScriptVar) <= Length(Buffer) then
  begin
    Move(Buffer[Ofs], ScriptVar, SizeOf(ScriptVar));
    Inc(Ofs, SizeOf(ScriptVar));
    ResourceText.Add(Format('%-*s[script variable: %d]', [RecLevel + 2, '',
      ScriptVar]))
  end
  else
    ResourceText.Add(Format('%-*s[script variable: (undefined value)]',
      [RecLevel + 2, '']));

  if Ofs < Length(Buffer) then
    ResourceText.Add(Format('%-*s[%d bytes of unknown data]', [RecLevel + 2, '',
      Length(Buffer) - Ofs]))
end;

procedure WriteSubtitle(RecLevel: Longword; Tag1: Byte;
  var ResourceText: TStringList);
var
  Ofs : Longint;
  Size: Word;
  Line: string;
begin
  Ofs := 0;
  Move(Buffer[Ofs], Size, SizeOf(Size));
  Inc(Ofs, SizeOf(Size));
  SetLength(Line, Size);
  if Size > 0 then
    Move(Buffer[Ofs], Line[1], Size);
  Inc(Ofs, Size);

  ResourceText.Add(Format('%-*s[subtitle: "%s"]', [RecLevel + 2, '', Line]));

  if Ofs < Length(Buffer) then
    ResourceText.Add(Format('%-*s[%d bytes of unknown data]', [RecLevel + 2, '',
      Length(Buffer) - Ofs]))
end;

procedure WriteModelDetails(RecLevel: Longword; Tag1: Byte;
  var ResourceText: TStringList);
var
  Ofs : Longint;
  Size: Word;
  Line: string;
begin
  Ofs := 0;
  Move(Buffer[Ofs], Size, SizeOf(Size));
  Inc(Ofs, SizeOf(Size));
  SetLength(Line, Size);
  if Size > 0 then
    Move(Buffer[Ofs], Line[1], Size);
  Inc(Ofs, Size);

  ResourceText.Add(Format('%-*s[model file: "%s"]', [RecLevel + 2, '', Line]));

  if Ofs < Length(Buffer) then
    ResourceText.Add(Format('%-*s[%d bytes of unknown data]', [RecLevel + 2, '',
      Length(Buffer) - Ofs]))
end;

procedure WriteLipsyncDetails(RecLevel: Longword; Tag1: Byte;
  var ResourceText: TStringList);
var
  i,
  Ofs,
  SyncSize,
  SequSize: Longint;
  Sync,
  Sequ    : string;
begin
  Ofs := 0;
  Move(Buffer[Ofs], SyncSize, SizeOf(SyncSize));
  Inc(Ofs, SizeOf(SyncSize));

  SetLength(Sync, SyncSize);
  for i := 0 to SyncSize - 1 do
    Sync[i + 1] := Chr(Buffer[i shl 3 + 4]);

  Inc(Ofs, SyncSize shl 3);

  Move(Buffer[Ofs], SequSize, SizeOf(SequSize));
  Inc(Ofs, SizeOf(SequSize));

  SetLength(Sequ, SequSize);
  if SequSize > 0 then
    Move(Buffer[Ofs], Sequ[1], SequSize);
  Inc(Ofs, SequSize);

  ResourceText.Add(Format('%-*s[sync: "%s"; load sequence: "%s"]',
    [RecLevel + 2, '', Sync, Sequ]));

  if Ofs < Length(Buffer) then
    ResourceText.Add(Format('%-*s[%d bytes of unknown data]', [RecLevel + 2, '',
      Length(Buffer) - Ofs]))
end;

procedure WriteStockSoundDetails(RecLevel: Longword; Tag1: Byte;
  var ResourceText: TStringList);
var
  Ofs,
  Duration,
  LocalID : Longint;
begin
  Ofs := 0;
  Move(Buffer[Ofs], Duration, SizeOf(Duration));
  Inc(Ofs, SizeOf(Duration));
  Move(Buffer[Ofs], LocalID, SizeOf(LocalID));
  Inc(Ofs, SizeOf(LocalID));

  ResourceText.Add(Format('%-*s[duration: %d ms; local ID: %d]', [RecLevel + 2,
    '', Duration, LocalID]));

  if Ofs < Length(Buffer) then
    ResourceText.Add(Format('%-*s[%d bytes of unknown data]', [RecLevel + 2, '',
      Length(Buffer) - Ofs]))
end;

procedure WriteTextureDetails(RecLevel: Longword; Tag1: Byte;
  var ResourceText: TStringList);
var
  Ofs : Longint;
  Size: Word;
  Line: string;
begin
  Ofs := 0;
  Move(Buffer[Ofs], Size, SizeOf(Size));
  Inc(Ofs, SizeOf(Size));
  SetLength(Line, Size);
  if Size > 0 then
    Move(Buffer[Ofs], Line[1], Size);
  Inc(Ofs, Size);

  ResourceText.Add(Format('%-*s[texture file: "%s"]', [RecLevel + 2, '',
    Line]));

  if Ofs < Length(Buffer) then
    ResourceText.Add(Format('%-*s[%d bytes of unknown data]', [RecLevel + 2, '',
      Length(Buffer) - Ofs]))
end;

procedure WriteXRCEntry(const XRCEntry: TXRCEntry; RecLevel: Longword;
  var ResourceText: TStringList);
var
  i: Longint;
begin
  with XRCEntry do
  begin
    if (DataSize > 0) or (NumChildren > 0) then
      ResourceText.Add(Format('%-*s%s {', [RecLevel, '', Name]))
    else
      ResourceText.Add(Format('%-*s%s', [RecLevel, '', Name]));

    case TypeID of
      $08: WriteObjectDetails(RecLevel, Tag1, ResourceText);
      $0b: WriteAnimationDetails(RecLevel, Tag1, ResourceText);
      $0d: WriteImageDetails(RecLevel, Tag1, ResourceText);
      $0f: WriteAnimScriptLineDetails(RecLevel, Tag1, ResourceText);
      $10: WriteSoundDetails(RecLevel, Tag1, ResourceText);
      $15: WriteScriptVarDetails(RecLevel, Tag1, ResourceText);
      $1d: WriteSubtitle(RecLevel, Tag1, ResourceText);
      $20: WriteModelDetails(RecLevel, Tag1, ResourceText);
      $23: WriteLipsyncDetails(RecLevel, Tag1, ResourceText);
      $24: WriteStockSoundDetails(RecLevel, Tag1, ResourceText);
      $26: WriteTextureDetails(RecLevel, Tag1, ResourceText)
    else
      if DataSize > 0 then
        ResourceText.Add(Format('%-*s[%d bytes of unknown data]', [RecLevel + 2,
          '', DataSize]))
    end;

    if NumChildren > 0 then
      for i := 0 to NumChildren - 1 do
        ReadXRC(RecLevel + 2, ResourceText);

    if (DataSize > 0) or (NumChildren > 0) then
      ResourceText.Add(Format('%-*s}', [RecLevel, '']))
  end
end;

procedure ReadXRC(RecLevel: Longword; var ResourceText: TStringList);
var
  XRCEntry: TXRCEntry;
begin
  with XRCEntry do
  begin
    FS_BlockRead(XRCFile, TypeID, 1);
    FS_BlockRead(XRCFile, Tag1, 1);
    FS_BlockRead(XRCFile, Tag2, 2);
    FS_BlockRead(XRCFile, NameLen, 2);
    SetLength(Name, NameLen);
    if NameLen > 0 then
      FS_BlockRead(XRCFile, Name[1], NameLen);
    FS_BlockRead(XRCFile, DataSize, 4);
    SetLength(Buffer, DataSize);
    if DataSize > 0 then
      FS_BlockRead(XRCFile, Buffer[0], DataSize);
    FS_BlockRead(XRCFile, NumChildren, 2);
    FS_BlockRead(XRCFile, Tag4, 2)
  end;

  WriteXRCEntry(XRCEntry, RecLevel, ResourceText)
end;

function ConvertXRC(const XRCName: string; var ResourceText: TStringList;
  const XRCCallback: TXRCCallback): Boolean;
begin
  Result := FS_OpenFile(XRCName, XRCFile);
  if not Result then
    Exit;

  ReadXRC(0, ResourceText);

  FS_CloseFile(XRCFile)
end;

end.

