{ File   : Main.pas
  Author : Deniz Oezmen
  Created: 2005-11-12
  Changed: 2010-09-08

  TLJView Main form control.
}
unit Main;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ComCtrls, StdCtrls, FileCtrl, ExtCtrls, ImgList, Menus, CommCtrl,
  OleCtrls, Buttons, ShellAPI, GLWin32Viewer, GLMisc, GLScene, GLCadencer,
  Animation, GLObjects, GLContext, Gauges;

type
  TMainForm = class(TForm)
    ResourceTreeView: TTreeView;
    StatusBar: TStatusBar;
    Splitter: TSplitter;
    ImageList: TImageList;
    MainMenu: TMainMenu;
    FileMenuItem: TMenuItem;
    ScrollBox: TScrollBox;
    Image: TImage;
    ProgressBar: TProgressBar;
    TextMemo: TMemo;
    LeftPanel: TPanel;
    DetailsComboBox: TComboBox;
    LeftBottomPanel: TPanel;
    SoundControlPanel: TPanel;
    PlaySpeedButton: TSpeedButton;
    PauseSpeedButton: TSpeedButton;
    StopSpeedButton: TSpeedButton;
    SoundTrackBar: TTrackBar;
    PlayTimer: TTimer;
    OptionsMenuItem: TMenuItem;
    AutoSoundMenuItem: TMenuItem;
    ShowMIPMapMenuItem: TMenuItem;
    ExitMenuItem: TMenuItem;
    SelectBaseDirMenuItem: TMenuItem;
    EditMenuItem: TMenuItem;
    ExportMenuItem: TMenuItem;
    ExportSaveDialog: TSaveDialog;
    ExportRawMenuItem: TMenuItem;
    AnimationControlPanel: TPanel;
    ModelComboBox: TComboBox;
    ModelLabel: TLabel;
    TextureComboBox: TComboBox;
    TextureLabel: TLabel;
    IncludeCheckBox: TCheckBox;
    Scene: TGLScene;
    GLSceneViewer: TGLSceneViewer;
    GLCamera: TGLCamera;
    GLCadencer: TGLCadencer;
    AntiAliasMenuItem: TMenuItem;
    AANoneMenuItem: TMenuItem;
    AA2xMenuItem: TMenuItem;
    AA4xMenuItem: TMenuItem;
    AAHighQualityMenuItem: TMenuItem;
    AABreakMenuItem: TMenuItem;
    AADefaultMenuItem: TMenuItem;
    SmoothRepetitionCheckBox: TCheckBox;
    AnimationProgressBar: TProgressBar;
    procedure ResourceTreeViewChange(Sender: TObject; Node: TTreeNode);
    procedure ResourceTreeViewExpanding(Sender: TObject; Node: TTreeNode;
      var AllowExpansion: Boolean);
    procedure ResourceTreeViewCollapsing(Sender: TObject; Node: TTreeNode;
      var AllowCollapse: Boolean);
    procedure PlaySpeedButtonClick(Sender: TObject);
    procedure PauseSpeedButtonClick(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure SplitterMoved(Sender: TObject);
    procedure DetailsComboBoxSelect(Sender: TObject);
    procedure StopSpeedButtonClick(Sender: TObject);
    procedure PlayTimerTimer(Sender: TObject);
    procedure SoundTrackBarChange(Sender: TObject);
    procedure AutoSoundMenuItemClick(Sender: TObject);
    procedure ShowMIPMapMenuItemClick(Sender: TObject);
    procedure ExitMenuItemClick(Sender: TObject);
    procedure SelectBaseDirMenuItemClick(Sender: TObject);
    procedure ExportMenuItemClick(Sender: TObject);
    procedure ExportRawMenuItemClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure IncludeCheckBoxClick(Sender: TObject);
    procedure ModelComboBoxSelect(Sender: TObject);
    procedure TextureComboBoxSelect(Sender: TObject);
    procedure GLCadencerProgress(Sender: TObject; const deltaTime,
      newTime: Double);
    procedure GLSceneViewerMouseMove(Sender: TObject; Shift: TShiftState;
      X, Y: Integer);
    procedure GLSceneViewerMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure AANoneMenuItemClick(Sender: TObject);
    procedure AA2xMenuItemClick(Sender: TObject);
    procedure AA4xMenuItemClick(Sender: TObject);
    procedure AAHighQualityMenuItemClick(Sender: TObject);
    procedure AADefaultMenuItemClick(Sender: TObject);
    procedure SmoothRepetitionCheckBoxClick(Sender: TObject);
  private
    procedure PopulateView(const BaseDir: string; ReScan: Boolean);
    procedure PopulateNode(const Node: TTreeNode);
    procedure SetImageSize(const Width, Height: Longint);
    procedure ShowImage(const Bitmap: TBitmap);
    procedure LoadSound(const SoundFileName: string);
    procedure SetSoundControls;
    procedure FillDetailsComboBox(const StringList: TStringList);
    procedure FillTextMemo(const StringList: TStringList);
    procedure StartExtAnimation(const AnimationFileName: string);
    procedure StartIntAnimation(const IntAnimation: TIntAnimation);
    procedure SetIntAnimationControls;
    procedure SetScene;
    procedure SizeControls;
    procedure ResetControls;
    procedure SetControls;
  public
    { Public-Deklarationen }
  end;

var
  MainForm: TMainForm;

implementation

uses
  SplWait, HndlType, ConvHndl, ExpHndl, TLJ_FS, Sound, Options;

{$R *.dfm}

var
  BaseDir    : string;
  HandlerData: THandlerData;
  ModelList,
  TextureList: TStringList;
  MouseX,
  MouseY     : Longint;

(* callback functions *********************************************************)

procedure ScanCallback(const Path: string);
begin
  with SplashWaitForm do
  begin
    MsgLabel.Caption :=
      MinimizeName(Path, Canvas, MsgPanel.Width - 8);
    MsgPanel.Repaint
  end;

  with MainForm do
  begin
    StatusBar.Panels[0].Text := MinimizeName(Path, Canvas,
      StatusBar.Panels[0].Width - 8);
    StatusBar.Repaint
  end
end;

procedure HandleCallback(const Frac: Double);
begin
  MainForm.ProgressBar.Position := Round(Frac * 100)
end;

(* general controls ***********************************************************)

procedure TMainForm.SizeControls;
var
  Rect: TRect;
begin
  // adapt text in panel
  if Assigned(ResourceTreeView.Selected) then
    if Assigned(ResourceTreeView.Selected.Data) then
      StatusBar.Panels[0].Text := MinimizeName(
        BaseDir + string(ResourceTreeView.Selected.Data^), StatusBar.Canvas,
        StatusBar.Panels[0].Width);

  // set image container size, centered
  SetImageSize(Image.Picture.Width, Image.Picture.Height);

  // set text viewer size
  if HandlerData.Intf_Type = Intf_SoundComment then
    TextMemo.Height := ScrollBox.ClientHeight - SoundControlPanel.Height - 1
  else
    TextMemo.Height := ScrollBox.ClientHeight;

  if HandlerData.Intf_Type = Intf_SoundComment then
    with TextMemo do
    begin
      ScrollBars := ssVertical;
      WordWrap := True
    end
  else
    with TextMemo do
    begin
      ScrollBars := ssBoth;
      WordWrap := False
    end;

  // set animation control size
  if Assigned(GLSceneViewer) then
  begin
    GLSceneViewer.Height := ScrollBox.ClientHeight -
      AnimationControlPanel.Height - 1;

    // only an educated guess ...
    GLCamera.SceneScale := GLSceneViewer.Height / 450;
    if GLSceneViewer.Width / 450 < GLCamera.SceneScale then
      GLCamera.SceneScale := GLSceneViewer.Width / 450
  end;

  // resize leftmost status bar panel
  StatusBar.Panels[0].Width := StatusBar.Width - 223;

  // set progress bar size
  StatusBar.Perform(SB_GETRECT, 1, Integer(@Rect));
  with Rect do
    ProgressBar.SetBounds(Left, Top, Right - Left, Bottom - Top)
end;

procedure TMainForm.ResetControls;
var
  Node: TTreeNode;
begin
  // menu
  ExportMenuItem.Enabled := False;
  ExportRawMenuItem.Enabled := False;

  // status bar
  Node := ResourceTreeView.Selected;

  if Assigned(Node) then
    if Node.Selected then
      if Assigned(Node.Data) then
      begin
        StatusBar.Panels[0].Text := MinimizeName(BaseDir + string(Node.Data^),
          StatusBar.Canvas, StatusBar.Panels[0].Width);
      end
      else
        StatusBar.Panels[0].Text := '';

  // progress bar
  ProgressBar.Position := 0;

  // image/drawing plane
  Image.Visible := False;
  Image.Picture.Bitmap.Width := 0;
  Image.Picture.Bitmap.Height := 0;
  SetImageSize(0, 0);

  // sound controls
  SoundControlPanel.Visible := False;
  PlayTimer.Enabled := False;
  UnassignSound;
  SetSoundControls;

  // animation controls
  AnimationControlPanel.Visible := False;
  GLSceneViewer.Visible := False;
  GLCadencer.Enabled := False;
  SetIntAnimationControls;
  SetScene;

  // text memo
  TextMemo.Visible := False;
  TextMemo.Lines.Clear;
  TextMemo.WordWrap := False;
  TextMemo.ScrollBars := ssBoth;

  // details combo box
  DetailsComboBox.Items.Clear;
  DetailsComboBox.Text := '';
  DetailsComboBox.Enabled := False
end;

procedure TMainForm.SetControls;
var
  Node: TTreeNode;
begin
  // menu
  Node := ResourceTreeView.Selected;

  if Assigned(Node) then
    if Node.Selected then
      if Assigned(Node.Data) then
      begin
        ExportMenuItem.Enabled :=
          ((Handle_GetExportType(HandlerData) <> Export_None) and
          (HandlerData.Intf_Type <> Intf_None));
        ExportRawMenuItem.Enabled := ExportMenuItem.Enabled
      end;

  // progress bar
  ProgressBar.Position := 0
end;

(* resource tree controls *****************************************************)

procedure TMainForm.PopulateNode(const Node: TTreeNode);
var
  i       : Longint;
  DirList,
  FileList: TFS_Entries;
  NewNode : TTreeNode;
  Desc    : PAnsiString;
begin
  // get directory content
  FS_GetDirList(DirList);
  FS_GetFileList(FileList);

  // add subdirectory nodes
  for i := 0 to High(DirList) do
  begin
    New(Desc);
    Desc^ := FS_GetCurrentDir + DirList[i].Name + '\';

    if DirList[i].FriendlyName <> '' then
      NewNode := ResourceTreeView.Items.AddChildObject(Node,
        DirList[i].FriendlyName, Desc)
    else
      NewNode := ResourceTreeView.Items.AddChildObject(Node,
        DirList[i].Name, Desc);

    NewNode.ImageIndex := Type_FolderClose;
    NewNode.SelectedIndex := Type_FolderClose;

    // recurse
    FS_ChangeDir(DirList[i].Name);
    PopulateNode(NewNode);
    FS_ChangeDir('..')
  end;

  // add file nodes
  for i := 0 to High(FileList) do
  begin
    New(Desc);
    Desc^ := FS_GetCurrentDir + FileList[i].Name;

    if FileList[i].FriendlyName <> '' then
      NewNode := ResourceTreeView.Items.AddChildObject(Node,
        FileList[i].FriendlyName, Desc)
    else
      NewNode := ResourceTreeView.Items.AddChildObject(Node,
        FileList[i].Name, Desc);

    NewNode.ImageIndex := Handle_GetFileType(FileList[i].Name);
    NewNode.SelectedIndex := NewNode.ImageIndex;

    // check for model or texture files for global list
    if NewNode.ImageIndex = Type_CIRModel then
      ModelList.Add(Desc^)
    else if NewNode.ImageIndex = Type_TMTexture then
      TextureList.Add(Desc^);
  end
end;

procedure TMainForm.PopulateView(const BaseDir: string; ReScan: Boolean);
var
  i       : Longint;
  FS_Entry: TFS_Entry;
  Desc    : PAnsiString;
begin
  if not ReScan then
  begin
    SplashWaitForm.Show;
    SplashWaitForm.BlendUp
  end;

  SplashWaitForm.Repaint;

  ResetControls;

  for i := 0 to ResourceTreeView.Items.Count - 1 do
    if Assigned(ResourceTreeView.Items[i].Data) then
    begin
      SetLength(PAnsiString(ResourceTreeView.Items[i].Data)^, 0);
      Dispose(ResourceTreeView.Items[i].Data);
      ResourceTreeView.Items[i].Data := nil
    end;
  ResourceTreeView.Items.Clear;

  // load file system
  if not FS_Init(BaseDir, ScanCallback) then
  begin
    ScanCallback('TLJ installation not found');
    SplashWaitForm.Close;
    Exit
  end;

  ScanCallback('Building resource tree ...');

  ResourceTreeView.Items.BeginUpdate;

  // add info for root node
  New(Desc);
  Desc^ := '\';
  FS_GetInfo(Desc^, FS_Entry);
  ResourceTreeView.Items.AddChild(nil, FS_Entry.FriendlyName);
  ResourceTreeView.Items[0].ImageIndex := Type_Root;
  ResourceTreeView.Items[0].SelectedIndex := Type_Root;
  ResourceTreeView.Items[0].Data := Desc;

  // clear global lists
  ModelList.Clear;
  TextureList.Clear;

  // everything else will be done by recursion
  PopulateNode(ResourceTreeView.Items[0]);

  SplashWaitForm.Close;

  ResourceTreeView.Items[0].Expand(False);
  ResourceTreeView.Items.EndUpdate
end;

procedure TMainForm.ResourceTreeViewChange(Sender: TObject; Node: TTreeNode);
begin
  if not Node.Selected then
    Exit;

  ResetControls;

  if not Assigned(Node.Data) then
    Exit;

  Screen.Cursor := crHourGlass;

  Handle_Select(Node.ImageIndex, string(Node.Data^), HandlerData,
    HandleCallback);

  case HandlerData.Intf_Type of
    Intf_Bitmap:
      with HandlerData do
        if Length(BitmapList) > 0 then
          ShowImage(BitmapList[0]);
    Intf_BitmapList:
      with HandlerData do
      begin
        FillDetailsComboBox(StringList);
        if Length(BitmapList) > 0 then
          ShowImage(BitmapList[0])
      end;
    Intf_Sound:
      with HandlerData do
        if SoundList.Count > 0 then
          LoadSound(SoundList[0]);
    Intf_SoundComment:
      with HandlerData do
      begin
        if SoundList.Count > 0 then
          LoadSound(SoundList[0]);
        FillTextMemo(StringList)
      end;
    Intf_TextDataList:
      with HandlerData do
        FillTextMemo(StringList);
    Intf_RawData:
      with HandlerData do
        FillTextMemo(StringList);
    Intf_ExtAnimation:
      with HandlerData do
        if ExtAnimationList.Count > 0 then
          StartExtAnimation(ExtAnimationList[0]);
    Intf_IntAnimation:
      with HandlerData do
        if Length(IntAnimationList) > 0 then
          StartIntAnimation(IntAnimationList[0])
  end;

  SetControls;

  Screen.Cursor := crDefault
end;

procedure TMainForm.ResourceTreeViewExpanding(Sender: TObject; Node: TTreeNode;
  var AllowExpansion: Boolean);
begin
  Node.ImageIndex := Handle_Expand(Node.ImageIndex);
  Node.SelectedIndex := Handle_Expand(Node.SelectedIndex)
end;

procedure TMainForm.ResourceTreeViewCollapsing(Sender: TObject;
  Node: TTreeNode; var AllowCollapse: Boolean);
begin
  Node.ImageIndex := Handle_Collapse(Node.ImageIndex);
  Node.SelectedIndex := Handle_Collapse(Node.SelectedIndex)
end;

(* image controls *************************************************************)

procedure TMainForm.SetImageSize(const Width, Height: Longint);
begin
  // set internal picture size
  Image.Picture.Bitmap.Width := Width;
  Image.Picture.Bitmap.Height := Height;

  // set image container size
  Image.Width := Width;
  Image.Height := Height;

  // crop container size
  if Image.Width < ScrollBox.ClientWidth then
    Image.Width := ScrollBox.ClientWidth;
  if Image.Height < ScrollBox.ClientHeight then
    Image.Height := ScrollBox.ClientHeight
end;

procedure TMainForm.ShowImage(const Bitmap: TBitmap);
begin
  if not Assigned(Bitmap) then
    Exit;

  // draw bitmap
  Image.Picture.Assign(Bitmap);
  Image.Transparent := Bitmap.Transparent;
  Image.Visible := True;

  SizeControls
end;

(* sound controls *************************************************************)

procedure TMainForm.LoadSound(const SoundFileName: string);
begin
  AssignSound(SoundFileName);

  // prepare controls
  SoundTrackBar.Max := GetSoundLength;
  SoundControlPanel.Visible := True;
  PlayTimer.Enabled := True;

  // play it, Sam
  if GetAutoSoundOption then
    PlaySpeedButtonClick(self)
end;

procedure TMainForm.SetSoundControls;
begin
  if not IsSoundAssigned then
  begin
    SoundTrackBar.Enabled := False;
    PlaySpeedButton.Enabled := False;
    PauseSpeedButton.Enabled := False;
    StopSpeedButton.Enabled := False;
    Exit
  end;

  SoundTrackbar.OnChange := nil;
  SoundTrackBar.Position := GetSoundPosition;
  SoundTrackBar.OnChange := SoundTrackBarChange;

  if SoundIsPlaying then
  begin
    SoundTrackBar.Enabled := True;
    PlaySpeedButton.Enabled := True;
    PauseSpeedButton.Enabled := True;
    StopSpeedButton.Enabled := True;
    PlaySpeedButton.Down := True
  end
  else if GetSoundPosition = 0 then
  begin
    SoundTrackBar.Enabled := True;
    PlaySpeedButton.Enabled := True;
    PauseSpeedButton.Enabled := False;
    StopSpeedButton.Enabled := False;
    StopSpeedButton.Down := True
  end
  else
  begin
    SoundTrackBar.Enabled := True;
    PlaySpeedButton.Enabled := True;
    PauseSpeedButton.Enabled := True;
    StopSpeedButton.Enabled := True;
    PauseSpeedButton.Down := True
  end
end;

procedure TMainForm.PlaySpeedButtonClick(Sender: TObject);
begin
  PlaySound
end;

procedure TMainForm.PauseSpeedButtonClick(Sender: TObject);
begin
  if SoundIsPlaying then
    PauseSound
  else
    PlaySound
end;

procedure TMainForm.StopSpeedButtonClick(Sender: TObject);
begin
  StopSound
end;

procedure TMainForm.PlayTimerTimer(Sender: TObject);
begin
  SetSoundControls
end;

procedure TMainForm.SoundTrackBarChange(Sender: TObject);
begin
  SetSoundPosition(SoundTrackBar.Position)
end;

(* details combo box controls *************************************************)

procedure TMainForm.FillDetailsComboBox(const StringList: TStringList);
begin
  DetailsComboBox.Enabled := (StringList.Count > 0);
  if not DetailsComboBox.Enabled then
    Exit;

  DetailsComboBox.Items.AddStrings(StringList);

  DetailsComboBox.ItemIndex := 0
end;

procedure TMainForm.DetailsComboBoxSelect(Sender: TObject);
begin
  with HandlerData do
    case Intf_Type of
      Intf_BitmapList:
        if DetailsComboBox.ItemIndex <= High(BitmapList) then
          ShowImage(BitmapList[DetailsComboBox.ItemIndex])
    end
end;

(* text memo controls *********************************************************)

procedure TMainForm.FillTextMemo(const StringList: TStringList);
begin
  with TextMemo do
  begin
    Lines.AddStrings(StringList);
    SelStart := 0;
    SelLength := 0
  end;

  TextMemo.Visible := True;
  SizeControls
end;

(* animation controls *********************************************************)

procedure TMainForm.StartExtAnimation(const AnimationFileName: string);
begin
  ShellExecute(Application.Handle, 'open', PChar(AnimationFileName), nil,
    PChar(GetCurrentDir), SW_SHOWNORMAL)
end;

procedure TMainForm.StartIntAnimation(const IntAnimation: TIntAnimation);
begin
  // prepare scene
  InitScene(Scene);

  // load animation
  AssignAnimation(IntAnimation);

  // initialize model if possible
  ModelComboBoxSelect(nil);

  // initialize textures if possible
  TextureComboBoxSelect(nil);

  // let's do this
  AnimationControlPanel.Visible := True;
  GLSceneViewer.Visible := True;
  GLCadencer.Reset;
  GLCadencer.Enabled := True
end;

procedure TMainForm.SetIntAnimationControls;
begin
  // clear last animation
  if IsAnimationAssigned then
  begin
    EraseAnimation;
    EraseModel;
    EraseTextureLib
  end;

  // update combo boxes
  IncludeCheckBoxClick(nil)
end;

procedure TMainForm.SetScene;
begin
  case GetAntiAliasOption of
    -1:
      begin
        GLSceneViewer.Buffer.AntiAliasing := aaDefault;
        AADefaultMenuItem.Checked := True
      end;
    0:
      begin
        GLSceneViewer.Buffer.AntiAliasing := aaNone;
        AANoneMenuItem.Checked := True
      end;
    2:
      begin
        if GetHighQualityOption then
          GLSceneViewer.Buffer.AntiAliasing := aa2xHQ
        else
          GLSceneViewer.Buffer.AntiAliasing := aa2x;
        AA2xMenuItem.Checked := True
      end;
    4:
      begin
        if GetHighQualityOption then
          GLSceneViewer.Buffer.AntiAliasing := aa4xHQ
        else
          GLSceneViewer.Buffer.AntiAliasing := aa4x;
        AA4xMenuItem.Checked := True
      end
  end;

  AAHighQualityMenuItem.Checked := GetHighQualityOption;

  // reload textures
  TextureComboBoxSelect(nil)
end;

procedure TMainForm.IncludeCheckBoxClick(Sender: TObject);
var
  i,
  MIdx,
  TIdx: Longint;
  Dir,
  CurM,
  CurT: string;
begin
  CurM := ModelComboBox.Text;
  CurT := TextureComboBox.Text;

  ModelComboBox.Items.Clear;
  TextureComboBox.Items.Clear;

  if IncludeCheckBox.Checked then
  begin
    ModelComboBox.Items.AddStrings(ModelList);
    TextureComboBox.Items.AddStrings(TextureList)
  end
  else
  begin
    if Assigned(ResourceTreeView.Selected) then
      if Assigned(ResourceTreeView.Selected.Data) then
      begin
        Dir := AnsiLowerCase(ExtractFilePath(
          string(ResourceTreeView.Selected.Data^)));

        for i := 0 to ModelList.Count - 1 do
          if AnsiLowerCase(ExtractFilePath(ModelList[i])) = Dir then
            ModelComboBox.Items.Add(ModelList[i]);
        for i := 0 to TextureList.Count - 1 do
          if AnsiLowerCase(ExtractFilePath(TextureList[i])) = Dir then
            TextureComboBox.Items.Add(TextureList[i])
      end
  end;

  MIdx := ModelComboBox.Items.IndexOf(CurM);
  if MIdx >= 0 then
    ModelComboBox.ItemIndex := MIdx
  else if ModelComboBox.Items.Count > 0 then
  begin
    ModelComboBox.ItemIndex := 0;
    ModelComboBoxSelect(nil)
  end;

  TIdx := TextureComboBox.Items.IndexOf(CurT);
  if TIdx >= 0 then
    TextureComboBox.ItemIndex := TIdx
  else if TextureComboBox.Items.Count > 0 then
  begin
    TextureComboBox.ItemIndex := 0;
    TextureComboBoxSelect(nil)
  end
end;

procedure TMainForm.ModelComboBoxSelect(Sender: TObject);
var
  HandlerData: THandlerData;
begin
  if ModelComboBox.Items.Count = 0 then
    Exit;

  if IsAnimationAssigned then
  begin
    Handle_AllocData(HandlerData);

    Screen.Cursor := crHourGlass;

    Handle_Select(Type_CIRModelRaw,
      ModelComboBox.Items[ModelComboBox.ItemIndex], HandlerData,
      HandleCallback);

    if HandlerData.Intf_Type = Intf_Model then
      AssignModel(HandlerData.ModelList[0]);

    Screen.Cursor := crDefault;

    Handle_FreeData(HandlerData)
  end;

  SetControls
end;

procedure TMainForm.TextureComboBoxSelect(Sender: TObject);
var
  HandlerData : THandlerData;
  MIPMapOption: Boolean;
begin
  if TextureComboBox.Items.Count = 0 then
    Exit;

  if IsAnimationAssigned then
  begin
    Handle_AllocData(HandlerData);

    // force loading without MIP maps
    MIPMapOption := GetShowMIPMapOption;
    SetShowMIPMapOption(False);

    Screen.Cursor := crHourGlass;

    Handle_Select(Type_TMTexture,
      TextureComboBox.Items[TextureComboBox.ItemIndex], HandlerData,
      HandleCallback);

    SetShowMIPMapOption(MIPMapOption);

    if HandlerData.Intf_Type = Intf_BitmapList then
      AssignTextureLib(HandlerData.BitmapList, HandlerData.StringList);

    Screen.Cursor := crDefault;

    Handle_FreeData(HandlerData)
  end;

  SetControls
end;

procedure TMainForm.GLCadencerProgress(Sender: TObject; const deltaTime,
  newTime: Double);
begin
  AnimationProgressBar.Position :=
    Round(UpdateAnimation(newTime, SmoothRepetitionCheckBox.Checked) *
    AnimationProgressBar.Max);

  // dummy cube is assumed to be object 2
  GLCamera.TargetObject := Scene.Objects[2];

  Scene.NotifyChange(nil)
end;

procedure TMainForm.GLSceneViewerMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
begin
  if Shift = [ssLeft] then
  begin
    GLCamera.MoveAroundTarget(MouseY - Y, MouseX - X);
    MouseY := Y;
    MouseX := X
  end
  else if Shift = [ssRight] then
  begin
    GLCamera.MoveTargetInEyeSpace(0, MouseX - X, Y - MouseY);
    MouseY := Y;
    MouseX := X
  end
end;

procedure TMainForm.GLSceneViewerMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  MouseX := X;
  MouseY := Y
end;

(* form controls **************************************************************)

procedure TMainForm.FormResize(Sender: TObject);
begin
{  if Application.Terminated then
    Exit;}

  SizeControls
end;

procedure TMainForm.FormCreate(Sender: TObject);
var
  NonClientMetrics: TNonClientMetrics;
begin
  // get system font information
  NonClientMetrics.cbSize := SizeOf(NonClientMetrics);
  SystemParametersInfo(SPI_GETNONCLIENTMETRICS, 0, @NonClientMetrics, 0);

  // scale form to compensate for font size differences
  if NonClientMetrics.lfMessageFont.lfHeight <> Font.Height then
    ScaleBy(NonClientMetrics.lfMessageFont.lfHeight, Font.Height);

  // set new font after (!) rescaling
  Font.Handle := CreateFontIndirect(NonClientMetrics.lfMessageFont);

  // connect progress bar and status bar
  ProgressBar.Parent := StatusBar
end;

procedure TMainForm.FormDestroy(Sender: TObject);
var
  i: Longint;
begin
  DeinitScene;

  GLSceneViewer.Free;
  Scene.Free;

  for i := 0 to ResourceTreeView.Items.Count - 1 do
    if Assigned(ResourceTreeView.Items[i].Data) then
    begin
      SetLength(PAnsiString(ResourceTreeView.Items[i].Data)^, 0);
      Dispose(ResourceTreeView.Items[i].Data);
      ResourceTreeView.Items[i].Data := nil
    end
end;

procedure TMainForm.FormShow(Sender: TObject);
begin
  SizeControls;

  AutoSoundMenuItem.Checked := GetAutoSoundOption;
  ShowMIPMapMenuItem.Checked := GetShowMIPMapOption;

  BaseDir := GetBaseDirOption;
  PopulateView(BaseDir, False)
end;

procedure TMainForm.SplitterMoved(Sender: TObject);
begin
  SizeControls
end;

(* menu controls **************************************************************)

procedure TMainForm.SelectBaseDirMenuItemClick(Sender: TObject);
var
  NewBaseDir: string;
begin
  if not SelectDirectory('Choose The Longest Journey installation directory:',
    '', NewBaseDir) then
    Exit;

  SetBaseDirOption(NewBaseDir);
  BaseDir := GetBaseDirOption;

  PopulateView(BaseDir, True)
end;

procedure TMainForm.AutoSoundMenuItemClick(Sender: TObject);
begin
  AutoSoundMenuItem.Checked := not AutoSoundMenuItem.Checked;
  SetAutoSoundOption(AutoSoundMenuItem.Checked)
end;

procedure TMainForm.ShowMIPMapMenuItemClick(Sender: TObject);
begin
  ShowMIPMapMenuItem.Checked := not ShowMIPMapMenuItem.Checked;
  SetShowMIPMapOption(ShowMIPMapMenuItem.Checked);

  // refresh
  with ResourceTreeView do
    if Assigned(Selected) then
      if Selected.ImageIndex = Type_TMTexture then
        ResourceTreeViewChange(self, Selected)
end;

procedure TMainForm.AADefaultMenuItemClick(Sender: TObject);
begin
  AADefaultMenuItem.Checked := True;
  SetAntiAliasOption(-1);

  SetScene
end;

procedure TMainForm.AANoneMenuItemClick(Sender: TObject);
begin
  AANoneMenuItem.Checked := True;
  SetAntiAliasOption(0);

  SetScene
end;

procedure TMainForm.AA2xMenuItemClick(Sender: TObject);
begin
  AA2xMenuItem.Checked := True;
  SetAntiAliasOption(2);

  SetScene
end;

procedure TMainForm.AA4xMenuItemClick(Sender: TObject);
begin
  AA4xMenuItem.Checked := True;
  SetAntiAliasOption(4);

  SetScene
end;

procedure TMainForm.AAHighQualityMenuItemClick(Sender: TObject);
begin
  AAHighQualityMenuItem.Checked := not AAHighQualityMenuItem.Checked;
  SetHighQualityOption(AAHighQualityMenuItem.Checked);

  SetScene
end;

procedure TMainForm.ExitMenuItemClick(Sender: TObject);
begin
  Close
end;

procedure TMainForm.ExportMenuItemClick(Sender: TObject);
var
  ExportName,
  ExportExt : string;
begin
  SetControls;

  case Handle_GetExportType(HandlerData) of
    Export_Single:
      begin
        // export a single file
        ExportName := Handle_GetExportDefaultName(HandlerData);
        ExportExt := ExtractFileExt(ExportName);

        with ExportSaveDialog do
        begin
          DefaultExt := ExportExt;
          Filter := Format('%s files (*%0:s)|*%0:s', [ExportExt]);
          FileName := ExportName;
          ForceCurrentDirectory := True;

          if Execute then
          begin
            Screen.Cursor := crHourGlass;

            if not Handle_Export(HandlerData, FileName, HandleCallback) then
              MessageDlg('Export failed. Ensure that you have sufficient righ' +
                'ts to write to the selected directory.', mtError, [mbOK], 0);

            Screen.Cursor := crDefault
          end
        end
      end;
    Export_Multiple:
      // export several files
      if SelectDirectory('Choose an export directory:', '', ExportName) then
      begin
        Screen.Cursor := crHourGlass;

        if not Handle_Export(HandlerData, ExportName, HandleCallback) then
          MessageDlg('Export failed. Ensure that you have sufficient rights t' +
            'o write to the selected directory.', mtError, [mbOK], 0);

        Screen.Cursor := crDefault
      end
  end;

  SetControls
end;

procedure TMainForm.ExportRawMenuItemClick(Sender: TObject);
var
  OldType,
  OldIntf: Longint;
begin
  // ugly little hack to force raw data being exported
  OldType := HandlerData.HandlerType;
  OldIntf := HandlerData.Intf_Type;

  HandlerData.HandlerType := Type_RawFile;
  HandlerData.Intf_Type := Intf_RawData;

  ExportMenuItemClick(Sender);

  HandlerData.HandlerType := OldType;
  HandlerData.Intf_Type := OldIntf
end;

procedure TMainForm.SmoothRepetitionCheckBoxClick(Sender: TObject);
begin
  GLCadencer.Reset
end;

initialization
  Handle_AllocData(HandlerData);
  ModelList := TStringList.Create;
  TextureList := TStringList.Create;

finalization
  Handle_FreeData(HandlerData);
  ModelList.Free;
  TextureList.Free;
  FS_DeInit

end.
