{ File   : XARC_FS.pas
  Author : Deniz Oezmen
  Created: 2005-11-13, based on XARCExt.dpr created 2005-10-23
  Changed: 2007-03-05

  XARC reader for The Longest Journey.
}
unit XARC_FS;

interface

type
  TXARCHeader  = packed record
                   Unknown,
                   NumFiles,
                   BaseOfs : Longint
                 end;

  TXARCEntry   = record
                   Name   : string;
                   Size,
                   Offset,            // not in file
                   Unknown: Longint
                 end;
  TXARCEntries = array of TXARCEntry;

  TXARC        = record
                   Name   : string;
                   Header : TXARCHeader;
                   Entries: TXARCEntries
                 end;

function ReadXARC(var XARC: TXARC): Boolean;
procedure ClearXARC(var XARC: TXARC);

implementation

uses
  SysUtils;

function ReadXARC(var XARC: TXARC): Boolean;
var
  i, j    : Longint;
  XARCFile: file;
  FileOfs : Longword;
  c       : Char;
begin
  Result := False;

  FileMode := fmOpenRead;
  AssignFile(XARCFile, XARC.Name);
  {$I-}
  Reset(XARCFile, 1);
  {$I+}
  if IOResult <> 0 then
    Exit;

  try
    BlockRead(XARCFile, XARC.Header, SizeOf(XARC.Header));

    SetLength(XARC.Entries, XARC.Header.NumFiles);
    FileOfs := XARC.Header.BaseOfs;

    for i := 0 to XARC.Header.NumFiles - 1 do
    begin
      with XARC.Entries[i] do
      begin
        SetLength(Name, 255);
        for j := 1 to 256 do
        begin
          BlockRead(XARCFile, c, 1);
          if c = #000 then
            Break
          else if j = 256 then
            Exit;

          Name[j] := c;
        end;
        SetLength(Name, j - 1);

        BlockRead(XARCFile, Size, 4);
        BlockRead(XARCFile, Unknown, 4);
        Offset := FileOfs;

        Inc(FileOfs, Size)
      end
    end
  finally
    Close(XARCFile)
  end;

  Result := True
end;

procedure ClearXARC(var XARC: TXARC);
begin
  Finalize(XARC.Entries)
end;

end.

