{ File    : $URL: file:///D:/SVN/svnDelphi/trunk/TLJ/TLJView/src/fs/XRC_FS.pas $
  Author  : $Author: Deniz Oezmen $
  Changed : $Date: 2020-11-30 19:36:23 +0100 (Mo, 30 Nov 2020) $
  Revision: $Rev: 586 $

  XRC reader for The Longest Journey.
}
unit XRC_FS;

interface                 

type
  TXRCEntry     = record
                    ID  : Byte;
                    Name: string;
                  end;

  TXRCLocations = array of TXRCEntry;

  TXRCFileEntry = record
                    Name,
                    FriendlyName: string;
                    ExtendedInfo: array of string;
                  end;

  TXRCFiles     = array of TXRCFileEntry;

  TXRC          = record
                    Name     : string;
                    BaseOfs,
                    Size     : Longint;
                    Header   : TXRCEntry;
                    Locations: TXRCLocations;
                    Files    : TXRCFiles;
                  end;

function ReadXRC(var XRC: TXRC): Boolean;
procedure ClearXRC(var XRC: TXRC);

implementation

uses
  SysUtils;

type
  TXRCRecord = record
                 TypeID,
                 Tag1       : Byte;
                 Tag2,
                 NameLen    : Word;
                 Name       : string;
                 DataSize   : Longint;
                 NumChildren,
                 Tag4       : Word;
               end;

  TBuffer    = array of Byte;

procedure ReadXRCRecord(var XRCFile: file; out XRCRecord: TXRCRecord;
  out DataBuffer: TBuffer);
begin
  with XRCRecord do
  begin
    BlockRead(XRCFile, TypeID, 1);
    BlockRead(XRCFile, Tag1, 1);
    BlockRead(XRCFile, Tag2, 2);
    BlockRead(XRCFile, NameLen, 2);
    SetLength(Name, NameLen);
    BlockRead(XRCFile, Name[1], NameLen);
    BlockRead(XRCFile, DataSize, 4);
    SetLength(DataBuffer, DataSize);
    if DataSize > 0 then
      BlockRead(XRCFile, DataBuffer[0], DataSize);
    BlockRead(XRCFile, NumChildren, 2);
    BlockRead(XRCFile, Tag4, 2)
  end;
end;

function ReadDataString(const DataBuffer: TBuffer; Offset: Longint;
  out DataString: string): Boolean;
var
  Size: Word;
begin
  Result := False;
  DataString := '';

  if Length(DataBuffer) < Offset + 2 then
    Exit;

  Move(DataBuffer[Offset], Size, 2);
  if Length(DataBuffer) < Offset + Size + 2 then
    Exit;

  SetLength(DataString, Size);
  if Size > 0 then
    Move(DataBuffer[Offset + 2], DataString[1], Size);

  Result := True
end;

procedure ProcessLocation(var XRCFile: file; const XRCRecord: TXRCRecord;
  var XRC: TXRC);
begin
  SetLength(XRC.Locations, Length(XRC.Locations) + 1);
  with XRC.Locations[High(XRC.Locations)] do
  begin
    ID := XRCRecord.Tag2;
    Name := XRCRecord.Name
  end
end;

procedure ProcessAnimation(var XRCFile: file; const XRCRecord: TXRCRecord;
  const DataBuffer: TBuffer; var XRC: TXRC);
var
  Ext,
  FileName: string;
begin
  if not ReadDataString(DataBuffer, 8, FileName) then
    Exit;

  //FIXME: only accept descriptions for SSS or BBB files -- others (.ani) are
  // confusing
  Ext := AnsiLowerCase(ExtractFileExt(FileName));

  if (Ext <> '.sss') and (Ext <> '.bbb') then
    Exit;

  SetLength(XRC.Files, Length(XRC.Files) + 1);
  with XRC.Files[High(XRC.Files)] do
  begin
    Name := FileName;
    FriendlyName := XRCRecord.Name
  end
end;

//FIXME: Not yet really useful
{procedure ProcessImage(var XRCFile: file; const XRCRecord: TXRCRecord;
  const DataBuffer: TBuffer; var XRC: TXRC);
var
  FileName: string;
begin
  if not ReadDataString(DataBuffer, 0, FileName) then
    Exit;

  SetLength(XRC.Files, Length(XRC.Files) + 1);
  with XRC.Files[High(XRC.Files)] do
  begin
    Name := FileName;
    FriendlyName := XRCRecord.Name
  end
end;}

procedure ProcessSound(var XRCFile: file; const XRCRecord: TXRCRecord;
  const DataBuffer: TBuffer; var XRC: TXRC);
var
  FileName: string;
begin
  if not ReadDataString(DataBuffer, 0, FileName) then
    Exit;

  SetLength(XRC.Files, Length(XRC.Files) + 1);
  with XRC.Files[High(XRC.Files)] do
  begin
    Name := FileName;
    FriendlyName := XRCRecord.Name
  end
end;

function ReadSoundFile(var XRCFile: file; const NumChildren: Longword;
  out SoundFileName, SoundName: string): Boolean;
var
  i         : Longint;
  XRCRecord : TXRCRecord;
  DataBuffer: TBuffer;
begin
  Result := False;
  SoundFileName := '';
  SoundName := '';

  for i := 0 to NumChildren - 1 do
  begin
    ReadXRCRecord(XRCFile, XRCRecord, DataBuffer);

    if XRCRecord.TypeID <> $10 then
      Continue;

    SoundName := XRCRecord.Name;
    Result := ReadDataString(DataBuffer, 0, SoundFileName);
    if Result then
      Break
  end
end;

procedure ProcessDialogue(var XRCFile: file; const XRCRecord: TXRCRecord;
  const DataBuffer: TBuffer; var XRC: TXRC);
var
  SoundFileName,
  SoundName,
  Subtitle     : string;
begin
  if not ReadSoundFile(XRCFile, XRCRecord.NumChildren, SoundFileName,
    SoundName) then
    Exit;

  SetLength(XRC.Files, Length(XRC.Files) + 1);
  with XRC.Files[High(XRC.Files)] do
  begin
    Name := SoundFileName;
    FriendlyName := SoundName;

    if ReadDataString(DataBuffer, 0, Subtitle) then
    begin
      SetLength(ExtendedInfo, 3);
      ExtendedInfo[0] := Format('[%s]', [XRCRecord.Name]);
      ExtendedInfo[1] := '';
      ExtendedInfo[2] := Subtitle
    end
    else
      Setlength(ExtendedInfo, 0)
  end
end;

//FIXME: Not yet really useful
{procedure ProcessModel(var XRCFile: file; const XRCRecord: TXRCRecord;
  const DataBuffer: TBuffer; var XRC: TXRC);
var
  FileName: string;
begin
  if not ReadDataString(DataBuffer, 0, FileName) then
    Exit;

  SetLength(XRC.Files, Length(XRC.Files) + 1);
  with XRC.Files[High(XRC.Files)] do
  begin
    Name := FileName;
    FriendlyName := XRCRecord.Name
  end
end;

procedure ProcessTexture(var XRCFile: file; const XRCRecord: TXRCRecord;
  const DataBuffer: TBuffer; var XRC: TXRC);
var
  FileName: string;
begin
  if not ReadDataString(DataBuffer, 0, FileName) then
    Exit;

  SetLength(XRC.Files, Length(XRC.Files) + 1);
  with XRC.Files[High(XRC.Files)] do
  begin
    Name := FileName;
    FriendlyName := XRCRecord.Name
  end
end;}

function ReadXRC(var XRC: TXRC): Boolean;
var
  XRCFile   : file;
  XRCRecord : TXRCRecord;
  DataBuffer: TBuffer;
begin
  Result := False;

  FileMode := fmOpenRead;
  AssignFile(XRCFile, XRC.Name);
  {$I-}
  Reset(XRCFile, 1);
  {$I+}
  if IOResult <> 0 then
    Exit;

  SetLength(XRC.Locations, 0);

  try
    Seek(XRCFile, XRC.BaseOfs);
    ReadXRCRecord(XRCFile, XRCRecord, DataBuffer);
    XRC.Header.ID := XRCRecord.Tag2;
    XRC.Header.Name := XRCRecord.Name;

    while FilePos(XRCFile) < XRC.BaseOfs + XRC.Size do
    begin
      ReadXRCRecord(XRCFile, XRCRecord, DataBuffer);

//FIXME: Deactivate some types -- friendly names are too confusing when viewed
      case XRCRecord.TypeID of
        $02: ProcessLocation(XRCFile, XRCRecord, XRC);
        $03: ProcessLocation(XRCFile, XRCRecord, XRC);
        $0b: ProcessAnimation(XRCFile, XRCRecord, DataBuffer, XRC);
//        $0d: ProcessImage(XRCFile, XRCRecord, DataBuffer, XRC);
        $10: ProcessSound(XRCFile, XRCRecord, DataBuffer, XRC);
        $1d: ProcessDialogue(XRCFile, XRCRecord, DataBuffer, XRC);
//        $20: ProcessModel(XRCFile, XRCRecord, DataBuffer, XRC);
//        $26: ProcessTexture(XRCFile, XRCRecord, DataBuffer, XRC)
      else
        Continue
      end
    end
  finally
    Close(XRCFile)
  end;

  Result := True
end;

procedure ClearXRC(var XRC: TXRC);
var
  i: Longint;
begin
  Finalize(XRC.Locations);

  for i := 0 to High(XRC.Files) do
    Finalize(XRC.Files[i].ExtendedInfo);
  Finalize(XRC.Files)
end;


end.
