{ File   : ConvHndl.pas
  Author : Deniz Oezmen
  Created: 2007-03-20, based on Handlers.pas created 2007-02-15
  Changed: 2007-08-30

  TLJView data conversion handler.
}
unit ConvHndl;

interface

uses
  HndlType;

function Handle_Expand(HandlerType: Longint): Longint;
function Handle_Collapse(HandlerType: Longint): Longint;
procedure Handle_Select(HandlerType: Longint; const Name: string;
  var HandlerData: THandlerData; HandleCallback: THandleCallback);

implementation

uses
  Windows, Graphics, Classes, SysUtils,
  TLJ_FS, Raw, XMG, OVS, ISN, TM, XRC, SSS, BBB, CIR, ANI, BIFF, Animation,
  Options;

const
  TempOGG = 'temp.ogg';
  TempWAV = 'temp.wav';
  TempSMK = 'temp.smk';
  TempBIK = 'temp.bik';

var
  OGGName,
  WAVName,
  SMKName,
  BIKName,
  TempDir: string;

function Handle_Expand(HandlerType: Longint): Longint;
begin
  Result := HandlerType;

  if HandlerType = Type_FolderClose then
    Result := Type_FolderOpen
end;

function Handle_Collapse(HandlerType: Longint): Longint;
begin
  Result := HandlerType;

  if HandlerType = Type_FolderOpen then
    Result := Type_FolderClose
end;

procedure Handle_RawFile(var HandlerData: THandlerData;
  HandleCallback: THandleCallback);
begin
  HandlerData.Intf_Type := Intf_None;

  if HandlerData.Name = '' then
    Exit;

  with HandlerData do
    if LoadRawAsHex(Name, HandlerData.StringList, HandleCallback) then
      Intf_Type := Intf_RawData
end;

procedure Handle_XMGPicture(var HandlerData: THandlerData;
  HandleCallback: THandleCallback);
var
  Bitmap: TBitmap;
begin
  HandlerData.Intf_Type := Intf_None;

  if HandlerData.Name = '' then
    Exit;

  with HandlerData do
  begin
    if not ConvertXMG(Name, Bitmap, HandleCallback) then
      Exit;

    SetLength(BitmapList, 1);
    BitmapList[0] := Bitmap;

    StringList.Clear;
    StringList.Add(Name);

    Intf_Type := Intf_Bitmap
  end
end;

procedure Handle_OVSSound(var HandlerData: THandlerData;
  HandleCallback: THandleCallback);
var
  i       : Longint;
  FS_Entry: TFS_Entry;
begin
  HandlerData.Intf_Type := Intf_None;

  if HandlerData.Name = '' then
    Exit;

  with HandlerData do
  begin
    if FileExists(OGGName) then
      DeleteFile(OGGName);

    if not CopyOVS(Name, OGGName, HandleCallback) then
      Exit;

    SoundList.Clear;
    SoundList.Add(OGGName);

    StringList.Clear;

    Intf_Type := Intf_Sound;

    if FS_GetInfo(Name, FS_Entry) then
    begin
      for i := 0 to High(FS_Entry.ExtendedInfo) do
        StringList.Add(FS_Entry.ExtendedInfo[i]);
      if Length(FS_Entry.ExtendedInfo) > 0 then
        Intf_Type := Intf_SoundComment
    end
  end
end;

procedure Handle_ISNSound(var HandlerData: THandlerData;
  HandleCallback: THandleCallback);
var
  i       : Longint;
  FS_Entry: TFS_Entry;
begin
  HandlerData.Intf_Type := Intf_None;

  if HandlerData.Name = '' then
    Exit;

  with HandlerData do
  begin
    if FileExists(WAVName) then
      DeleteFile(WAVName);

    if not ConvertISN(Name, WAVName, HandleCallback) then
      Exit;

    SoundList.Clear;
    SoundList.Add(WAVName);

    StringList.Clear;

    Intf_Type := Intf_Sound;

    if FS_GetInfo(Name, FS_Entry) then
    begin
      for i := 0 to High(FS_Entry.ExtendedInfo) do
        StringList.Add(FS_Entry.ExtendedInfo[i]);
      if Length(FS_Entry.ExtendedInfo) > 0 then
        Intf_Type := Intf_SoundComment
    end
  end
end;

procedure Handle_TMTexture(var HandlerData: THandlerData;
  HandleCallback: THandleCallback);
var
  i        : Longint;
  TMEntries: TTMEntries;
begin
  HandlerData.Intf_Type := Intf_None;

  if HandlerData.Name = '' then
    Exit;

  with HandlerData do
  begin
    if not ConvertTM(Name, TMEntries, GetShowMIPMapOption, HandleCallback) then
      Exit;

    SetLength(BitmapList, Length(TMEntries));
    StringList.Clear;
    for i := 0 to High(TMEntries) do
    begin
      BitmapList[i] := TMEntries[i].Bitmap;
      StringList.Add(TMEntries[i].Name)
    end;

    Intf_Type := Intf_BitmapList
  end;
end;

procedure Handle_XRCResource(var HandlerData: THandlerData;
  HandleCallback: THandleCallback);
begin
  HandlerData.Intf_Type := Intf_None;

  if HandlerData.Name = '' then
    Exit;

  with HandlerData do
    if ConvertXRC(Name, HandlerData.StringList, HandleCallback) then
      Intf_Type := Intf_TextDataList
end;

procedure Handle_SSSAnimation(var HandlerData: THandlerData;
  HandleCallback: THandleCallback);
begin
  HandlerData.Intf_Type := Intf_None;

  if HandlerData.Name = '' then
    Exit;

  with HandlerData do
  begin
    if FileExists(SMKName) then
      DeleteFile(SMKName);

    if not CopySSS(Name, SMKName, HandleCallback) then
      Exit;

    ExtAnimationList.Clear;
    ExtAnimationList.Add(SMKName);

    StringList.Clear;

    Intf_Type := Intf_ExtAnimation
  end
end;

procedure Handle_BBBAnimation(var HandlerData: THandlerData;
  HandleCallback: THandleCallback);
begin
  HandlerData.Intf_Type := Intf_None;

  if HandlerData.Name = '' then
    Exit;

  with HandlerData do
  begin
    if FileExists(BIKName) then
      DeleteFile(BIKName);

    if not CopyBBB(Name, BIKName, HandleCallback) then
      Exit;

    ExtAnimationList.Clear;
    ExtAnimationList.Add(BIKName);

    StringList.Clear;

    Intf_Type := Intf_ExtAnimation
  end
end;

procedure Handle_CIRModel(var HandlerData: THandlerData;
  HandleCallback: THandleCallback);
begin
  HandlerData.Intf_Type := Intf_None;

  if HandlerData.Name = '' then
    Exit;

  with HandlerData do
    if LoadCIRAsText(Name, HandlerData.StringList, HandleCallback) then
      Intf_Type := Intf_TextDataList
end;

procedure Handle_ANIAnimation(var HandlerData: THandlerData;
  HandleCallback: THandleCallback);
var
  IntAnimation: TIntAnimation;
begin
  HandlerData.Intf_Type := Intf_None;

  if HandlerData.Name = '' then
    Exit;

  with HandlerData do
  begin
    if not LoadANI(Name, IntAnimation, HandleCallback) then
      Exit;

    SetLength(IntAnimationList, 1);
    IntAnimationList[0] := IntAnimation;

    Intf_Type := Intf_IntAnimation
  end
end;

procedure Handle_BIFFObject(var HandlerData: THandlerData;
  HandleCallback: THandleCallback);
begin
  HandlerData.Intf_Type := Intf_None;

  if HandlerData.Name = '' then
    Exit;

  with HandlerData do
    if ConvertBIFF(Name, HandlerData.StringList, HandleCallback) then
      Intf_Type := Intf_TextDataList
end;

procedure Handle_CIRModelRaw(var HandlerData: THandlerData;
  HandleCallback: THandleCallback);
var
  Model: TModel;
begin
  HandlerData.Intf_Type := Intf_None;

  if HandlerData.Name = '' then
    Exit;

  with HandlerData do
  begin
    if not LoadCIR(Name, Model, HandleCallback) then
      Exit;

    SetLength(ModelList, 1);
    ModelList[0] := Model;

    Intf_Type := Intf_Model
  end
end;

procedure Handle_Select(HandlerType: Longint; const Name: string;
  var HandlerData: THandlerData; HandleCallback: THandleCallback);
begin
  Handle_EraseData(HandlerData);

  HandlerData.Name := Name;
  HandlerData.HandlerType := HandlerType;

  case HandlerType of
    Type_RawFile:
      Handle_RawFile(HandlerData, HandleCallback);
    Type_XMGPicture:
      Handle_XMGPicture(HandlerData, HandleCallback);
    Type_OVSSound:
      Handle_OVSSound(HandlerData, HandleCallback);
    Type_ISNSound:
      Handle_ISNSound(HandlerData, HandleCallback);
    Type_TMTexture:
      Handle_TMTexture(HandlerData, HandleCallback);
    Type_XRCResource:
      Handle_XRCResource(HandlerData, HandleCallback);
    Type_SSSAnimation:
      Handle_SSSAnimation(HandlerData, HandleCallback);
    Type_BBBAnimation:
      Handle_BBBAnimation(HandlerData, HandleCallback);
    Type_CIRModel:
      Handle_CIRModel(HandlerData, HandleCallback);
    Type_ANIAnimation:
      Handle_ANIAnimation(HandlerData, HandleCallback);
    Type_BIFFObject:
      Handle_BIFFObject(HandlerData, HandleCallback);
    Type_CIRModelRaw:
      Handle_CIRModelRaw(HandlerData, HandleCallback)
  end
end;

initialization
  SetLength(TempDir, MAX_PATH);
  SetLength(TempDir, GetTempPath(MAX_PATH, PChar(TempDir)));

  OGGName := TempDir + TempOGG;
  WAVName := TempDir + TempWAV;
  SMKName := TempDir + TempSMK;
  BIKName := TempDir + TempBIK;

finalization
  DeleteFile(OGGName);
  DeleteFile(WAVName);
  DeleteFile(BIKName);
  DeleteFile(SMKName)

end.
