{ File   : HndlType.pas
  Author : Deniz Oezmen
  Created: 2007-03-20, based on Handlers.pas created 2007-02-15
  Changed: 2007-09-21

  TLJView data type handler.
}
unit HndlType;

interface

uses
  Graphics, Classes, Animation;

const
  // handler types
  Type_None         = -1;
  Type_Root         = 0;
  Type_FolderClose  = 1;
  Type_FolderOpen   = 2;
  Type_RawFile      = 3;
  Type_XMGPicture   = 4;
  Type_OVSSound     = 5;
  Type_ISNSound     = 6;
  Type_TMTexture    = 7;
  Type_XRCResource  = 8;
  Type_SSSAnimation = 9;
  Type_BBBAnimation = 10;
  Type_CIRModel     = 11;
  Type_ANIAnimation = 12;
  Type_BIFFObject   = 13;
  Type_CIRModelRaw  = 14;

  // interface types
  Intf_None         = -1;
  Intf_Bitmap       = 0;
  Intf_BitmapList   = 1;
  Intf_Sound        = 2;
  Intf_SoundList    = 3;
  Intf_SoundComment = 4;
  Intf_TextData     = 5;
  Intf_TextDataList = 6;
  Intf_RawData      = 7;
  Intf_ExtAnimation = 8;
  Intf_IntAnimation = 9;
  Intf_Model        = 10;
  Intf_ModeList     = 11;

type
  THandleCallback = procedure(const Frac: Double);

  TBitmapList = array of TBitmap;

  TIntAnimationList = array of TIntAnimation;

  TModelList = array of TModel;

  THandlerData = record
                   Name            : string;
                   HandlerType     : Longint;
                   Intf_Type       : Longint;
                   BitmapList      : TBitmapList;
                   SoundList       : TStringList;
                   ExtAnimationList: TStringList;
                   IntAnimationList: TIntAnimationList;
                   ModelList       : TModelList;
                   StringList      : TStringList
                 end;

procedure Handle_AllocData(var HandlerData: THandlerData);
procedure Handle_EraseData(var HandlerData: THandlerData);
procedure Handle_FreeData(var HandlerData: THandlerData);
function Handle_GetFileType(const FileName: string): Longint;

implementation

uses
  SysUtils;

procedure Handle_AllocData(var HandlerData: THandlerData);
begin
  with HandlerData do
  begin
    Name := '';
    HandlerType := Type_None;
    Intf_Type := Intf_None;
    SetLength(BitmapList, 0);
    SoundList := TStringList.Create;
    ExtAnimationList := TStringList.Create;
    SetLength(IntAnimationList, 0);
    SetLength(ModelList, 0);
    StringList := TStringList.Create
  end
end;

procedure Handle_EraseData(var HandlerData: THandlerData);
var
  i: Longint;
begin
  with HandlerData do
  begin
    Name := '';
    HandlerType := Type_None;
    Intf_Type := Intf_None;

    for i := 0 to High(BitmapList) do
    begin
      BitmapList[i].FreeImage;
      BitmapList[i].Free;
      BitmapList[i] := nil
    end;

    Finalize(BitmapList);
    SetLength(BitmapList, 0);

    SoundList.Clear;
    ExtAnimationList.Clear;

    for i := 0 to High(IntAnimationList) do
      EraseAnimation(IntAnimationList[i]);

    Finalize(IntAnimationList);
    SetLength(IntAnimationList, 0);

    for i := 0 to High(ModelList) do
      EraseModel(ModelList[i]);

    Finalize(ModelList);
    SetLength(ModelList, 0);
    
    StringList.Clear
  end
end;

procedure Handle_FreeData(var HandlerData: THandlerData);
begin
  Handle_EraseData(HandlerData);

  with HandlerData do
  begin
    BitmapList := nil;
    SoundList.Free;
    SoundList := nil;
    ExtAnimationList.Free;
    ExtAnimationList := nil;
    IntAnimationList := nil;
    ModelList := nil;
    StringList.Free;
    StringList := nil
  end
end;

function Handle_GetFileType(const FileName: string): Longint;
var
  Ext: string;
begin
  Ext := AnsiLowerCase(ExtractFileExt(FileName));

  if Ext = '.xmg' then
    Result := Type_XMGPicture
  else if Ext = '.ovs' then
    Result := Type_OVSSound
  else if (Ext = '.isn') or (Ext = '.iss') or (Ext = '.ssn') or
    (Ext = '.sn') then
    Result := Type_ISNSound
  else if Ext = '.tm' then
    Result := Type_TMTexture
  else if Ext = '.xrc' then
    Result := Type_XRCResource
  else if Ext = '.sss' then
    Result := Type_SSSAnimation
  else if Ext = '.bbb' then
    Result := Type_BBBAnimation
  else if Ext = '.cir' then
    Result := Type_CIRModel
  else if Ext = '.ani' then
    Result := Type_ANIAnimation
  else if Ext = '.biff' then
    Result := Type_BIFFObject
  else
    Result := Type_RawFile
end;

end.
