{ File   : Options.pas
  Author : Deniz Oezmen
  Created: 2007-03-05
  Changed: 2007-08-30

  Option .ini management.
}
unit Options;

interface

function GetBaseDirOption: string;
procedure SetBaseDirOption(const BaseDir: string);
function GetAutoSoundOption: Boolean;
procedure SetAutoSoundOption(const AutoSound: Boolean);
function GetShowMIPMapOption: Boolean;
procedure SetShowMIPMapOption(const ShowMIPMap: Boolean);
function GetAntiAliasOption: Longint;
procedure SetAntiAliasOption(const AntiAlias: Longint);
function GetHighQualityOption: Boolean;
procedure SetHighQualityOption(const HighQuality: Boolean);

implementation

uses
  SysUtils, IniFiles,
  ShlHndl;

const
  AppDataFolder = 'TLJView';
  IniName       = '\TLJView.ini';

  isOptions     = 'Options';
  isBaseDir     = isOptions;
  isAutoSound   = isOptions;
  isShowMIPMap  = isOptions;
  isAntiAlias   = isOptions;
  isHighQuality = isOptions;

  inBaseDir     = 'BaseDir';
  inAutoSound   = 'AutoSound';
  inShowMIPMap  = 'ShowMIPMap';
  inAntiAlias   = 'AntiAlias';
  inHighQuality = 'HighQuality';

var
  Enabled : Boolean;
  DataPath: string;
  IniFile : TIniFile;

  idBaseDir    : string  = '';
  idAutoSound  : Boolean = True;
  idShowMIPMap : Boolean = True;
  idAntiAlias  : Longint = -1;
  idHighQuality: Boolean = False;

function GetBaseDirOption: string;
begin
  Result := idBaseDir;

  if not Enabled then
    Exit;

  Result := IniFile.ReadString(isBaseDir, inBaseDir, idBaseDir);
  if Result <> '' then
    while Result[Length(Result)] = '\' do
      Result := Copy(Result, 1, Length(Result) - 1)
end;

procedure SetBaseDirOption(const BaseDir: string);
begin
  idBaseDir := BaseDir;

  if not Enabled then
    Exit;

  IniFile.WriteString(isBaseDir, inBaseDir, BaseDir)
end;

function GetAutoSoundOption: Boolean;
begin
  Result := idAutoSound;

  if not Enabled then
    Exit;

  Result := IniFile.ReadBool(isAutoSound, inAutoSound, idAutoSound)
end;

procedure SetAutoSoundOption(const AutoSound: Boolean);
begin
  idAutoSound := AutoSound;

  if not Enabled then
    Exit;

  IniFile.WriteBool(isAutoSound, inAutoSound, AutoSound)
end;

function GetShowMIPMapOption: Boolean;
begin
  Result := idShowMIPMap;

  if not Enabled then
    Exit;

  Result := IniFile.ReadBool(isShowMIPMap, inShowMIPMap, idShowMIPMap)
end;

procedure SetShowMIPMapOption(const ShowMIPMap: Boolean);
begin
  idShowMIPMap := ShowMIPMap;

  if not Enabled then
    Exit;

  IniFile.WriteBool(isShowMIPMap, inShowMIPMap, ShowMIPMap)
end;

function GetAntiAliasOption: Longint;
begin
  Result := idAntiAlias;

  if not Enabled then
    Exit;

  Result := IniFile.ReadInteger(isAntiAlias, inAntiAlias, idAntiAlias);
  if not (Result in [0, 2, 4]) and (Result <> -1) then
    Result := idAntiAlias
end;

procedure SetAntiAliasOption(const AntiAlias: Longint);
begin
  if not (AntiAlias in [0, 2, 4]) and (AntiAlias <> -1) then
    Exit;

  idAntiAlias := AntiAlias;

  if not Enabled then
    Exit;

  IniFile.WriteInteger(isAntiAlias, inAntiAlias, AntiAlias)
end;

function GetHighQualityOption: Boolean;
begin
  Result := idHighQuality;

  if not Enabled then
    Exit;

  Result := IniFile.ReadBool(isHighQuality, inHighQuality, idHighQuality)
end;

procedure SetHighQualityOption(const HighQuality: Boolean);
begin
  idHighQuality := HighQuality;

  if not Enabled then
    Exit;

  IniFile.WriteBool(isHighQuality, inHighQuality, HighQuality)
end;

initialization
  Enabled := True;

  DataPath := GetDataPath(AppDataFolder);
  if not DirectoryExists(DataPath) then
    Enabled := CreateDir(DataPath);

  if Enabled then
    IniFile := TIniFile.Create(DataPath + IniName)

finalization
  if Enabled then
    IniFile.Free

end.
