{ File   : XARCExt.dpr
  Author : Deniz Oezmen
  Created: 2005-10-23
  Changed: 2005-10-23

  The Longest Journey XARC extractor.
}
program XARCExt;

{$APPTYPE CONSOLE}

uses
  SysUtils;

type
  TXARCHeader = packed record
                  Unknown,
                  NumFiles,
                  BaseOfs : Longword
                end;

  TXARCEntry  = packed record
                  Name   : string;
                  Size,
                  Offset,            // not in file
                  Unknown: Longword
                end;

var
  XARCHeader: TXARCHeader;
  XARCFile  : file;

function OpenXARCFile(var XARCFile: file; XARCName: string): Boolean;
begin
  FileMode := fmShareDenyWrite;
  AssignFile(XARCFile, XARCName);

  {$I-}
  Reset(XARCFile, 1);
  {$I+}
  OpenXARCFile := (IOResult = 0)
end;

function ReadHeader: Boolean;
var
  BytesRead: Longword;
begin
  BlockRead(XARCFile, XARCHeader, SizeOf(TXARCHeader), BytesRead);
  ReadHeader := BytesRead = SizeOf(TXARCHeader)
end;

procedure ExtractFiles;
var
  i, j,
  FileOfs : Longword;
  OutFile : file;
  c       : Char;
  Buffer  : array[0..32767] of Byte;
  FileList: array of TXARCEntry;
begin
  SetLength(FileList, XARCHeader.NumFiles);

  FileOfs := XARCHeader.BaseOfs;

  for i := 1 to XARCHeader.NumFiles do
  begin
    with FileList[i - 1] do
    begin
      Name := '';
      repeat
        BlockRead(XARCFile, c, 1);
        if c <> #000 then
          Name := Name + c
      until c = #000;

      BlockRead(XARCFile, Size, 4);
      BlockRead(XARCFile, Unknown, 4);
      Offset := FileOfs;

      Inc(FileOfs, Size)
    end
  end;

  for i := 0 to High(FileList) do
  begin
    with FileList[i] do
    begin
      Writeln('Extracting ', Name, ' @ ', Offset, ' (', Size, ' bytes)');

      Seek(XARCFile, Offset);

      FileMode := fmOpenReadWrite;
      AssignFile(OutFile, Name);
      Rewrite(OutFile, 1);

      for j := 1 to Size div SizeOf(Buffer) do
      begin
        BlockRead(XARCFile, Buffer, SizeOf(Buffer));
        BlockWrite(OutFile, Buffer, SizeOf(Buffer))
      end;
      BlockRead(XARCFile, Buffer, Size mod SizeOf(Buffer));
      BlockWrite(OutFile, Buffer, Size mod SizeOf(Buffer));

      CloseFile(OutFile)
    end
  end
end;

procedure ParseFiles(FilePara: string);
var
  DirPos,
  i        : Byte;
  FileName,
  Dir,
  XARCName  : string;
  SearchRec: TSearchRec;
  PFilePara: PChar;
begin
  GetMem(PFilePara, Length(FilePara) + 1);
  StrPCopy(PFilePara, FilePara);

  // get the directory information from FilePara
  DirPos := 0;
  for i := 1 to Length(FilePara) do
    if FilePara[i] in [':', '\'] then
      DirPos := i;

  Dir := Copy(FilePara, 1, DirPos);

  FileMode := fmOpenRead;

  // search for all files matching FilePara
  if FindFirst(PFilePara, faAnyFile - faDirectory - faVolumeID,
    SearchRec) = 0 then
  begin
    repeat
      FileName := SearchRec.Name;
      XARCName := Dir + FileName;

      // if this file is accessible, analyze it
      if OpenXARCFile(XARCFile, XARCName) then
        if ReadHeader then
        begin
          Writeln('Extracting from ', XARCName, ' ...');
          ExtractFiles;
          Close(XARCFile)
        end
        else
          Writeln(XARCName, ' is not a valid XARC file.')
      else
        Writeln(XARCName, ' not accessible')
    until FindNext(SearchRec) <> 0
  end
  else
    Writeln(FilePara, ' not found')
end;

begin
  if ParamCount = 1 then
    ParseFiles(ParamStr(1))
  else
    Writeln('Incorrect number of parameters.')
end.

